from lsdensities.core import hlt_matrix
from lsdensities.transform import ft_mp


def test_Smat_convergence():
    S_list = []
    T_list = []
    upper_bound = int(500)

    Sinf = hlt_matrix(upper_bound, alpha=0.0, type="EXP")
    fixed_Sinf = float(Sinf[3, 2])
    print("$S_{3,2}_{inf}$ = ", fixed_Sinf)

    for i in range(100, upper_bound, 100):
        tmax = int(i / 2 - 1)
        Scalc = hlt_matrix(tmax, type="COSH", alpha=0.0, T=i)
        S_list.append(float(Scalc[3, 2]))
        T_list.append(i)
        print("$S_{3,2}$ = ", float(Scalc[3, 2]))

    for i in range(1, len(S_list)):
        assert abs(S_list[i] - fixed_Sinf) < abs(
            S_list[i - 1] - fixed_Sinf
        ), f"Failed at T={T_list[i]}"

    assert abs(S_list[-1] - fixed_Sinf) < 1e-2


def test_ft_convergence():
    ft_list = []
    T_list = []
    sigma = 0.4
    fixed_t = 3
    fixed_e = 2
    upper_bound = int(200)

    ftinf = ft_mp(fixed_e, fixed_t, sigma, alpha=0, type="EXP")
    print("$f_{3}_{inf}$ = ", float(ftinf))

    for i in range(10, upper_bound, 4):
        ftcalc = ft_mp(fixed_e, fixed_t, sigma, alpha=0, type="COSH", T=i)
        ft_list.append(ftcalc)
        T_list.append(i)
        print("$f_{3}$ = ", float(ftcalc))

    for i in range(1, len(ft_list)):
        assert abs(ft_list[i] - ftinf) < abs(
            ft_list[i - 1] - ftinf
        ), f"Failed at T={T_list[i]}"

    assert abs(ft_list[-1] - ftinf) < 1e-6



import numpy as np
from numpy import linalg as LA
import math
import scipy.linalg as sp_linalg
from mpmath import mp


def kronecker_fp(a, b):
    if a == b:
        return 1
    if a != b:
        return 0


def halfnorm_fp(e, s):  # int_0^inf dE exp{(-e-e0)^2/2s^2}
    res_ = math.erf(e / (np.sqrt(2) * s))
    res_ += 1
    res_ *= s
    res_ *= np.sqrt(np.pi / 2)
    return res_


def gauss_fp(x, x0, sigma, norm="Full"):
    if sigma == 0:
        return kronecker_fp(x, x0)
    if norm == "Full" or norm == "full":
        return (np.exp(-((x - x0) ** 2) / (2 * (sigma**2)))) / (
            sigma * math.sqrt(2 * math.pi)
        )
    if norm == "None" or norm == "none":
        return np.exp(-((x - x0) ** 2) / (2 * sigma**2))
    if norm == "Half" or norm == "half":
        return (np.exp(-((x - x0) ** 2) / (2 * sigma**2))) / halfnorm_fp(x0, sigma)


def cauchy(k, sigma_, omega_):
    aux = omega_ - k
    aux = aux * aux + sigma_ * sigma_
    aux = sigma_ / aux
    return aux


def norm2_fp(matrix):  # for square matrices only
    assert matrix.shape[0] == matrix.shape[1]
    return LA.norm(matrix) / np.sqrt(matrix.shape[0])


def choelesky_invert_scipy(
    in_,
):  # invert positive definite matrix. wee faster than numpy
    _L, _lower = sp_linalg.cho_factor(in_)
    out_ = sp_linalg.cho_solve((_L, _lower), np.eye(in_.shape[0]))
    return out_


def norm2_mp(matrix):  # for square matrices only
    assert matrix.cols == matrix.rows
    return mp.norm(matrix) / mp.sqrt(matrix.cols)


def invert_matrix_ge(mp_matrix):
    matrix = mp_matrix.tolist()
    # Verify if the matrix is square
    n = len(matrix)
    if n != len(matrix[0]):
        raise ValueError("Matrix must be square.")

    # Create augmented matrix [A | I]
    augmented_matrix = [row + [0] * n for row in matrix]
    for i in range(n):
        augmented_matrix[i][n + i] = 1

    # Perform Gaussian elimination with partial pivoting
    for i in range(n):
        # Find pivot row
        pivot_row = max(range(i, n), key=lambda j: abs(augmented_matrix[j][i]))

        # Swap rows if necessary
        if pivot_row != i:
            augmented_matrix[i], augmented_matrix[pivot_row] = (
                augmented_matrix[pivot_row],
                augmented_matrix[i],
            )

        pivot = augmented_matrix[i][i]

        # Divide pivot row by pivot element
        for j in range(n * 2):
            augmented_matrix[i][j] /= pivot

        # Subtract multiples of pivot row from other rows
        for k in range(n):
            if k != i:
                factor = augmented_matrix[k][i]
                for j in range(n * 2):
                    augmented_matrix[k][j] -= factor * augmented_matrix[i][j]

    # Extract inverse matrix from augmented matrix
    inverse_matrix = [row[n:] for row in augmented_matrix]
    mp_inverse_matrix = mp.matrix(inverse_matrix)

    return mp_inverse_matrix



from mpmath import mp, mpf
from .utils.rhoUtils import LogMessage
from .transform import ft_mp
from .core import a0_scalar

#   Functionals A, B and W


def gAg(smat, gt, estar, alpha, params):
    tmax = params.tmax
    term1 = mpf(0)
    vt_ = mp.matrix(tmax, 1)
    for t in range(tmax):
        vt_[t] = 0
        for r in range(tmax):
            aux_ = mp.fmul(gt[r], smat[t, r])
            vt_[t] = mp.fadd(vt_[t], aux_)
        aux_ = mp.fmul(gt[t], vt_[t])
        term1 = mp.fadd(term1, aux_)
    #   term 1 = g^T S g

    term2 = a0_scalar(
        e=estar,
        sigma=params.mpsigma,
        alpha=alpha,
        e0=params.mpe0,
        ker_type=params.kerneltype,
    )
    #   term 2 = A0

    term3 = mpf(0)
    for t in range(tmax):
        aux_ = mp.fmul(
            ft_mp(
                e=estar,
                t=mpf(t + 1),
                sigma_=params.mpsigma,
                alpha=alpha,
                e0=params.mpe0,
                type=params.periodicity,
                T=params.time_extent,
                ker_type=params.kerneltype,
            ),
            gt[t],
        )
        term3 = mp.fadd(term3, aux_)
    term3 = mp.fmul(mpf(2), term3)
    #   term 3 = 2 sum_t f_t g_t

    res = mp.fsub(term1, term3)
    res = mp.fadd(res, term2)
    return res  #    term1 + term2 - term3


def gAgA0(smat, gt, estar, alpha, params, a0):
    tmax = params.tmax
    term1 = mpf(0)
    vt_ = mp.matrix(tmax, 1)
    for t in range(tmax):
        vt_[t] = 0
        for r in range(tmax):
            aux_ = mp.fmul(gt[r], smat[t, r])
            vt_[t] = mp.fadd(vt_[t], aux_)
        aux_ = mp.fmul(gt[t], vt_[t])
        term1 = mp.fadd(term1, aux_)
        term1 = mp.fdiv(term1, a0)
    #   term 1 = g^T S g / A0

    term2 = mpf(1)
    #   term 2 = A0 / A0

    term3 = 0
    for t in range(tmax):
        aux_ = mp.fmul(
            ft_mp(
                estar,
                mpf(t + 1),
                params.mpsigma,
                params.mpalpha,
                params.mpe0,
                params.periodicity,
                params.time_extent,
                ker_type=params.kerneltype,
            ),
            gt[t],
        )
        term3 = mp.fadd(term3, aux_)
    term3 = mp.fmul(mpf(2), term3)
    term3 = mp.fdiv(term3, a0)
    #   term 3 = 2 sum_t f_t g_t / A0

    res = mp.fsub(term1, term3)
    res = mp.fadd(res, term2)
    return res  # term1 + term2 - term3


def gBg(gt, bmat, bnorm):
    res = mpf(0)
    tmax = bmat.cols
    vt_ = mp.matrix(tmax, 1)
    for t in range(tmax):
        vt_[t] = 0
        for r in range(tmax):
            aux_ = mp.fmul(gt[r], bmat[r, t])
            vt_[t] = mp.fadd(aux_, vt_[t])
        aux_ = mp.fmul(gt[t], vt_[t])
        res = mp.fadd(res, aux_)
    res = mp.fdiv(res, bnorm)
    #   res = g B g / bnorm
    return res


def gWg(smat, cov, gt, estar, mplambda, a0_, bnorm, params, verbose=False):
    aterm = gAg(smat, gt, estar, params)
    aterm = mp.fdiv(aterm, a0_)
    if verbose is True:
        print(
            LogMessage(),
            "ABW :::",
            "Energy {:2.5f}".format(float(estar)),
            "Lambda {:2.3f}".format(float(mplambda)),
            "A/A0 = ",
            float(aterm),
        )
    bterm = gBg(gt, cov, bnorm)
    if verbose is True:
        print(
            LogMessage(),
            "ABW :::",
            "Energy {:2.5f}".format(float(estar)),
            "Lambda {:2.3f}".format(float(mplambda)),
            "B/Bnorm = ",
            float(bterm),
        )
    scale = mp.fsub(mpf(1), mplambda)
    wterm = mp.fmul(aterm, scale)  #   (1-l) A/A0
    bterm = mp.fmul(mplambda, bterm)  #   l B/Bnorm
    wterm = mp.fadd(wterm, bterm)  #  ( 1-l) A/A0 + l B/Bnorm
    return wterm



from lsdensities.utils.rhoUtils import (
    Inputs,
)
from mpmath import mp, mpf
from lsdensities.core import hlt_matrix
from lsdensities.transform import coefficients_ssd, get_ssd_scalar
from lsdensities.utils.rhoMath import gauss_fp


def run(time_extent):
    parameters = Inputs()
    parameters.time_extent = time_extent
    parameters.kerneltype = "FULLNORMGAUSS"
    parameters.periodicity = "EXP"
    parameters.sigma = 0.25
    parameters.prec = 128
    peak = 1
    energy = 0.5
    parameters.assign_values()

    lattice_correlator = mp.matrix(parameters.tmax, 1)
    lattice_covariance = mp.matrix(parameters.tmax)

    for t in range(parameters.tmax):  # mock data
        lattice_correlator[t] = mp.exp(-mpf(t + 1) * mpf(str(peak)))
        lattice_covariance[t, t] = lattice_correlator[t] * 0.02

    regularised_matrix = hlt_matrix(parameters.tmax, alpha=0)
    matrix_inverse = regularised_matrix ** (-1)

    coeff = coefficients_ssd(
        matrix_inverse,
        parameters,
        energy,
        alpha=0,
    )

    result = get_ssd_scalar(
        coeff,
        lattice_correlator,
        parameters,
    )

    true_value = gauss_fp(peak, energy, parameters.sigma, norm="full")

    diff = float(abs(result - true_value))
    print("time extent : ", time_extent, "diff = ", diff)

    if time_extent >= 32:
        assert diff < 1e-6
    elif time_extent < 32 and time_extent > 8:
        assert diff < 1e-1

    return float(diff)


def test_convergence():
    diff8 = run(8)
    diff16 = run(16)
    diff32 = run(32)
    diff48 = run(48)
    diff64 = run(64)

    assert diff8 > diff16
    assert diff16 > diff32
    assert diff32 > diff48
    assert diff48 > diff64


test_convergence()



from lsdensities.utils.rhoUtils import generate_seed, Inputs, init_precision


def rng_seeding():
    init_precision(128)
    par = Inputs()

    # Typical parameter setting
    par.time_extent = 32
    par.sigma = 0.25
    par.massNorm = 0.33
    par.Ne = 10
    par.emin = 0.3
    par.emax = 2.2

    seed1 = generate_seed(par)
    # Changing parameter
    par.Ne = 11

    seed2 = generate_seed(par)

    # Coming back to first setting
    par.Ne = 10
    seed3 = generate_seed(par)

    assert seed1 != seed2 and seed1 == seed3



from mpmath import mp, mpf
from .core import ft_mp, gte
from .utils.rhoStat import average_1d_mpmatrix, average_2d_mpmatrix


def coefficients_ssd(matrix, params, estar, alpha):
    """
    Computes the coefficients spanning the smeared spectral density
        gt = hlt_matrix * ft_mp

    Operation is performed for a single energy "estar"
    """
    gt = mp.matrix(params.tmax, 1)
    for i in range(params.tmax):
        for j in range(params.tmax):
            gt[i] += matrix[i, j] * ft_mp(
                e=estar,
                t=mpf(j + 1),
                sigma_=params.mpsigma,
                alpha=mpf(alpha),
                e0=params.mpe0,
                type=params.periodicity,
                T=params.time_extent,
                ker_type=params.kerneltype,
            )
    return gt


def get_ssd_scalar(gt, corr, params):
    """
    Computes smeared spectral density rho = sum_t g(t) c(t) at fixed energy
    Scalar version: the operation is performed at a single energy

    :param gt: mp.matrix len(params.tmax)
    :param corr: mp.matrix len(params.tmax)
    :param params: instance of Inputs class
    :return: mpf(float)
    """
    rho = 0
    for i in range(params.tmax):
        aux_ = mp.fmul(gt[i], corr[i])
        rho = mp.fadd(rho, aux_)
    return rho


def get_ssd_vector(gt, corr, params):
    """
    Computes smeared spectral density rho = sum_t g(t) c(t) at fixed energy
    Vector version: the operation is performed on a single vector of dimension par.tmax
    for a vector of energies

    :param gt: mp.matrix len(params.Ne, params.tmax)
    :param corr: mp.matrix len(params.tmax)
    :param params: instance of Inputs class
    :return: mp.matrix(params.Ne)
    """
    rho = mp.matrix(params.Ne, 1)
    for e in range(params.Ne):
        rho[e] = 0
        for i in range(params.tmax):
            aux_ = mp.fmul(gt[e, i], corr[i])
            rho[e] = mp.fadd(rho[e], aux_)
    return rho


def get_ssd_averaged_vector(gt, corr_type, params):
    """
    Computes smeared spectral density rho = sum_t g(t) c(t) at fixed energy.
    Averaged version: the operation is performed on a vector of correlators, corresponding to different statistical samples. Result is averaged.
    Vector version: the operation is performed at an array of energies.

    :param gt: mp.matrix len(params.Ne, params.tmax)
    :param corr_type: instance of Obs type
    :param params: instance of Inputs class
    :return: [mp.matrix(params.Ne), mp.matrix(params.Ne)] corresponding to avg and std
    """
    rhob = mp.matrix(params.Ne, params.num_boot)
    for b in range(params.num_boot):
        y = corr_type.sample[b][:]
        for e in range(params.Ne):
            rhob[e, b] = 0
            for i in range(params.tmax):
                aux_ = mp.fmul(gt[e, i], y[i])
                rhob[e, b] = mp.fadd(rhob[e, b], aux_)
    return average_2d_mpmatrix(rhob)


def get_ssd_averaged_scalar(gt, corr_samples, params):
    """
    Computes smeared spectral density rho = sum_t g(t) c(t) at fixed energy
    Averaged version: the operation is performed on a vector of correlators, corresponding to different statistical samples. Result is averaged.
    Scalar version: the operation is performed at a single energy.

    :param gt: mp.matrix len(params.Ne, params.tmax)
    :param corr_samples: mp.matrix of dimensions (params.num_boot, params.tmax).
    :param params: instance of Inputs class
    :return: [mpf(float), mpf(float)] corresponding to avg and std
    """
    rhob = mp.matrix(params.num_boot, 1)
    for b in range(params.num_boot):
        y = corr_samples[b, :]
        rhob[b] = 0
        for i in range(params.tmax):
            aux_ = mp.fmul(gt[i], y[i])
            rhob[b] = mp.fadd(rhob[b], aux_)
    return average_1d_mpmatrix(rhob)


def combine_fMf_scalar(gt, params, estar, alpha):
    """
    Computes f Minv f = g_t * f_t
    Scalar version: operation is performed at a single energy "estar"
    """
    out_ = 0
    for i in range(params.tmax):
        aux_ = mp.fmul(
            gt[i],
            ft_mp(
                estar,
                mpf(i + 1),
                sigma_=params.mpsigma,
                alpha=alpha,
                e0=params.mpe0,
                type=params.periodicity,
                T=params.time_extent,
                ker_type=params.kerneltype,
            ),
        )
        out_ = mp.fadd(out_, aux_)
    return out_


def combine_base_scalar(gt, params, estar):
    """
    Computes sum_t g(t) exp(-tE) or its periodic generalisation
    Scalar version: operation is performed at a single energy "estar".
    """
    out_ = 0
    for i in range(params.tmax):
        aux_ = mp.fmul(
            gt[i],
            gte(
                T=params.time_extent,
                t=mpf(i + 1),
                e=mpf(str(estar)),
                periodicity=params.periodicity,
            ),
        )
        out_ = mp.fadd(out_, aux_)
    return out_


def y_combine_sample_Eslice_mp_ToFile(file, ht_sliced, mpmatrix, params):
    rhob = mp.matrix(params.num_boot, 1)
    with open(file, "w") as output:
        for b in range(params.num_boot):
            y = mpmatrix[b, :]
            rhob[b] = 0
            for i in range(params.tmax):
                aux_ = mp.fmul(ht_sliced[i], y[i])
                rhob[b] = mp.fadd(rhob[b], aux_)
            print(b, float(rhob[b]), file=output)
        # print(LogMessage(), "rho[e] +/- stat ", float(average_1d_mpmatrix(rhob)[0]), (float(average_1d_mpmatrix(rhob)[1])))
    return average_1d_mpmatrix(rhob)


def combine_likelihood(minv, params, mpcorr):
    """
    :param minv: mp.matrix
    :param params: instance of Inputs class
    :param mpcorr: mp.matrix of len params.tmax
    :return: mpf(float)
    """
    out_ = 0
    aux = mp.matrix(params.tmax, 1)
    for i in range(params.tmax):
        aux[i] = 0
        for j in range(params.tmax):
            aux[i] += minv[i, j] * mpcorr[j]
        out_ += aux[i] * mpcorr[i]
    return out_



from ..utils.rhoUtils import Obs, LogMessage
import argparse
import numpy as np

#   Usage:
#       from correlatorUtils import effective_mass
#       effmass = effective_mass(corr, par, type='EXP')
#       effmass.plot(logscale=False)
#       print(effmass.central, '±', effmass.err)


def effective_mass(corr, par, type="COSH"):
    th = int(par.time_extent / 2)
    thm = th - 1
    mass = Obs(T=thm, nms=par.num_boot, is_resampled=True, tmax=thm)
    if type == "COSH":
        mass.sample[:, :] = np.arccosh(
            (corr.sample[:, 2 : th + 1] + corr.sample[:, 0 : th - 1])
            / (2 * corr.sample[:, 1:th])
        )
    elif type == "EXP":
        mass.sample[:, :] = -np.log(corr.sample[:, 1:th] / corr.sample[:, 0 : th - 1])
    else:
        raise ValueError("Invalid type specified. Only COSH and EXP are allowed.")

    mass.evaluate()
    return mass


def foldPeriodicCorrelator(corr, par, is_resampled=False):
    assert par.periodicity == "COSH"
    halfT = int(par.time_extent / 2)
    foldedCorr = Obs(
        T=halfT + 1, tmax=par.tmax, nms=par.num_samples, is_resampled=is_resampled
    )
    for n in range(par.num_samples):
        foldedCorr.sample[n, 0] = corr.sample[n, 0]
        for t in range(1, halfT + 1):
            foldedCorr.sample[n, t] = (
                corr.sample[n, t] + corr.sample[n, par.time_extent - t]
            ) / 2

    return foldedCorr


def symmetrisePeriodicCorrelator(corr, par):
    assert par.periodicity == "COSH"
    symmCorr = Obs(
        T=corr.T, tmax=corr.tmax, nms=corr.nms, is_resampled=corr.is_resampled
    )
    for n in range(par.num_samples):
        symmCorr.sample[n, 0] = corr.sample[n, 0]
        symmCorr.sample[n, int(symmCorr.T / 2)] = corr.sample[n, int(symmCorr.T / 2)]
        for t in range(1, int(symmCorr.T / 2)):
            symmCorr.sample[n, t] = (corr.sample[n, t] + corr.sample[n, corr.T - t]) / 2
            symmCorr.sample[n, symmCorr.T - t] = symmCorr.sample[n, t]

    return symmCorr


class InputsCorrelatorAnalysis:
    def __init__(
        self,
        time_extent: int = 0,
        datapath=".",
        outdir=".",
        num_boot: int = 0,
        num_samples: int = 0,
    ):
        if not isinstance(time_extent, int):
            raise TypeError("time_extent must be an integer")
        if not isinstance(num_samples, int):
            raise TypeError("num_samples must be an integer")
        if not isinstance(num_boot, int):
            raise TypeError("num_boot must be an integer")
        self.time_extent = time_extent
        self.datapath = datapath
        self.outdir = outdir
        self.num_boot = num_boot
        self.num_samples = num_samples
        self.periodicity = "EXP"
        self.tmax = 0

    def report(self):
        print(LogMessage(), "Init ::: ", "Reading file:", self.datapath)
        print(LogMessage(), "Init ::: ", "Output directory:", self.outdir)
        print(LogMessage(), "Init ::: ", "Time extent:", self.time_extent)
        print(LogMessage(), "Init ::: ", "Samples :", self.num_samples)
        print(LogMessage(), "Init ::: ", "Bootstrap samples :", self.num_boot)


def parseArgumentCorrelatorAnalysis():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-datapath",
        metavar="DataPile",
        type=str,
        help="Path to data file",
        required=True,
    )
    parser.add_argument(
        "--outdir", metavar="OutputDirectory", help="Directory for output", default="."
    )
    parser.add_argument(
        "--nboot",
        metavar="BootstrapSampleSize",
        type=int,
        help="Number of bootstrap samples. Default=300",
        default=300,
    )
    parser.add_argument(
        "--periodicity",
        type=str,
        help="Accepted stirngs are 'EXP' or 'COSH', depending on the correlator being periodic or open.",
        default="EXP",
    )
    args = parser.parse_args()
    return args



from lsdensities.utils.rhoMath import norm2_mp
from lsdensities.core import hlt_matrix
from lsdensities.utils.rhoUtils import LogMessage
from mpmath import mp


def test_mp_invert():
    print(LogMessage(), " Initialising...")
    mp.dps = 64
    print(LogMessage(), " Binary precision in bit: ", mp.prec)
    print(LogMessage(), " Approximate decimal precision: ", mp.dps)
    tmax = 15

    S = hlt_matrix(tmax, alpha=0)

    invS = S ** (-1)

    identity = S * invS

    shouldbeone = norm2_mp(identity)

    print(LogMessage(), "norm( S Sinv ) - 1 = ", shouldbeone - 1)
    print(LogMessage(), "Target decimal precision was", mp.dps)

    assert shouldbeone - 1 < mp.dps / 2



from lsdensities.utils.rhoUtils import LogMessage
from mpmath import mp, mpf


def test_init_precision():
    print(LogMessage(), " Initialising...")
    mp.dps = 10
    print(LogMessage(), " Binary precision in bit: ", mp.prec)
    print(LogMessage(), " Approximate decimal precision: ", mp.dps)
    a = mpf(1 / 3)
    assert a - 0.3333333333 < 1e-10



import numpy as np
import random
from mpmath import mp
import lsdensities.utils.rhoUtils as u
from .rhoUtils import LogMessage, Inputs
import multiprocessing as multiprocessing
from typing import List


def bootstrap_obstype_fp(par_, in_obs_type):
    """
    bootstrap the sample attribute of in_obs_type
    """
    from .rhoUtils import ranvec

    out_ = np.zeros((par_.num_boot, par_.time_extent))
    randv = np.zeros(par_.num_samples)
    for b in range(0, par_.num_boot):
        randv = ranvec(randv, par_.num_samples, 0, par_.num_samples)
        for i in range(0, par_.time_extent):
            out_[b][i] = 0
            for j in range(0, par_.num_samples):
                out_[b][i] += in_obs_type.sample[int(randv[j])][i]
            out_[b][i] /= par_.num_samples
    return out_


def bootstrap_sample_fp(par_, in_sample):
    """
    bootstrap in_sample
    """
    from .rhoUtils import ranvec

    out_ = np.zeros((par_.num_boot, par_.time_extent))
    randv = np.zeros(par_.num_samples)
    for b in range(0, par_.num_boot):
        randv = ranvec(randv, par_.num_samples, 0, par_.num_samples)
        for i in range(0, par_.time_extent):
            out_[b][i] = 0
            for j in range(0, par_.num_samples):
                out_[b][i] += in_sample[int(randv[j])][i]
            out_[b][i] /= par_.num_samples
    return out_


def parallel_bootstrap_sample_fp(par_, in_, out_, start, end, seed, is_folded=False):
    """
    parallelisable version of bootstrap_sample_fp
    """
    import lsdensities.utils.rhoUtils

    random.seed(seed)
    randv = np.zeros(par_.num_samples)
    if is_folded is False:
        for b in range(start, end):
            randv = lsdensities.utils.rhoUtils.ranvec(
                randv, par_.num_samples, 0, par_.num_samples
            ).astype(int)
            for i in range(par_.time_extent):
                out_[b][i] = np.mean(in_[randv[:], i])
    if is_folded is True:
        for b in range(start, end):
            randv = lsdensities.utils.rhoUtils.ranvec(
                randv, par_.num_samples, 0, par_.num_samples
            ).astype(int)
            for i in range(int(par_.time_extent / 2) + 1):
                out_[b][i] = np.mean(in_[randv[:], i])


def average_2d_mpmatrix(in_, bootstrap=True):
    """
    average 2D mpmatrix on the second axis, for each value of the first axis
    returns a 2D mpmatrix to be interpreted as two 1D mpmatrix:
    return: mp_array_mean, mp_array_std
    """
    xlen_ = in_.rows
    samplesize_ = in_.cols
    out_ = mp.matrix(xlen_, 2)
    for x in range(xlen_):
        out_[x, 0] = 0
        out_[x, 1] = 0
        for b in range(samplesize_):
            out_[x, 0] = mp.fadd(out_[x, 0], in_[x, b])
        out_[x, 0] = mp.fdiv(out_[x, 0], samplesize_)
        for b in range(samplesize_):
            aux_ = mp.fsub(in_[x, b], out_[x, 0])
            aux_ = mp.fmul(aux_, aux_)
            out_[x, 1] = mp.fadd(out_[x, 1], aux_)
        out_[x, 1] = mp.fdiv(out_[x, 1], samplesize_)
        out_[x, 1] = mp.sqrt(out_[x, 1])
        if bootstrap is False:
            aux_ = mp.sqrt(samplesize_)
            out_[x, 1] = mp.fdiv(out_[x, 1], aux_)
    return out_


def average_1d_mpmatrix(in_, bootstrap=True):
    """
    average 1D mpmatrix
    """
    if in_.rows == 1:
        samplesize_ = in_.cols
    if in_.cols == 1:
        samplesize_ = in_.rows
    out_ = mp.matrix(2, 1)
    out_[0] = 0
    out_[1] = 0
    for b in range(samplesize_):
        out_[0] = mp.fadd(out_[0], in_[b])
    out_[0] = mp.fdiv(out_[0], samplesize_)
    for b in range(samplesize_):
        aux_ = mp.fsub(in_[b], out_[0])
        aux_ = mp.fmul(aux_, aux_)
        out_[1] = mp.fadd(out_[1], aux_)
    out_[1] = mp.fdiv(out_[1], samplesize_)
    out_[1] = mp.sqrt(out_[1])
    if bootstrap is False:
        aux_ = mp.sqrt(samplesize_)
        out_[1] = mp.fdiv(out_[1], aux_)
    return out_


def resample(input_corr, par, parallelise=False):
    if parallelise is True:
        if par.periodicity == "EXP":
            output_corr = u.Obs(
                T=par.time_extent, tmax=par.tmax, nms=par.num_boot, is_resampled=True
            )
            resample = ParallelBootstrapLoop(par, input_corr.sample, is_folded=False)
        if par.periodicity == "COSH":
            output_corr = u.Obs(
                T=input_corr.T,
                tmax=input_corr.tmax,
                nms=par.num_boot,
                is_resampled=True,
            )
            resample = ParallelBootstrapLoop(par, input_corr.sample, is_folded=True)
        output_corr.sample = resample.run()

    else:
        output_corr = u.Obs(
            T=input_corr.T,
            tmax=input_corr.tmax,
            nms=par.num_boot,
            is_resampled=True,
        )
        output_corr.sample = bootstrap_obstype_fp(par, input_corr)

    output_corr.evaluate()
    return output_corr


class ParallelBootstrapLoop:
    def __init__(self, par: Inputs, input_sample: np.ndarray, is_folded=False):
        self.par = par
        self.looplen = par.num_boot
        self.vlen = par.time_extent
        self.inputsample = input_sample
        self.num_processes = multiprocessing.cpu_count()
        self.chunk_size = self.looplen // self.num_processes
        if self.looplen % self.num_processes != 0:
            self.chunk_size += 1
        self.out_array = multiprocessing.Array("d", self.looplen * self.vlen)
        self.out_ = np.frombuffer(self.out_array.get_obj()).reshape(
            (self.looplen, self.vlen)
        )
        self.processes: List[multiprocessing.Process] = []
        self.is_folded = is_folded

    def run(self) -> np.ndarray:
        for i in range(self.num_processes):
            start = i * self.chunk_size
            end = min(start + self.chunk_size, self.looplen)
            if self.is_folded is False:
                process = multiprocessing.Process(
                    target=parallel_bootstrap_sample_fp,
                    args=(
                        self.par,
                        self.inputsample,
                        self.out_,
                        start,
                        end,
                        random.randint(0, 2 ** (32) - 1),
                    ),
                )
            if self.is_folded is True:
                process = multiprocessing.Process(
                    target=parallel_bootstrap_sample_fp,
                    args=(
                        self.par,
                        self.inputsample,
                        self.out_,
                        start,
                        end,
                        random.randint(0, 2 ** (32) - 1),
                        self.is_folded,
                    ),
                )
            try:
                process.start()
            except Exception as e:
                print(f"Failed to start process {i}: {e}")
                self.terminate_all_processes()
                raise
            self.processes.append(process)
        print(LogMessage(), "Bootstrap ::: Running parallel loop")
        for process in self.processes:
            try:
                process.join()
            except Exception as e:
                print(f"Failed to join process {process}: {e}")
                self.terminate_all_processes()
                raise
        print(LogMessage(), "Bootstrap ::: End loop, joining processes")
        return self.out_

    def terminate_all_processes(self) -> None:
        for process in self.processes:
            if process.is_alive():
                process.terminate()
            process.join()



import rhoUtils as u
from ..rhoUtils import LogMessage
from ..utils.rhoParser import parseArgumentCorrelatorAnalysis
from ..utils.rhoParallelUtils import ParallelBootstrapLoop
from .correlatorUtils import InputsCorrelatorAnalysis


def main():
    print(LogMessage(), "Initialising")
    args = parseArgumentCorrelatorAnalysis()
    par = InputsCorrelatorAnalysis(
        datapath=args.datapath, outdir=args.outdir, num_boot=args.nboot
    )

    #   Reading datafile, storing correlator
    rawcorr, par.time_extent, par.num_samples = u.read_datafile(par.datapath)
    par.report()

    #   Here is the correlator
    rawcorr.evaluate()

    #   Here is the resampling
    corr = u.Obs(par.time_extent, par.num_boot, is_resampled=True)
    resample = ParallelBootstrapLoop(par, rawcorr.sample)
    corr.sample = resample.run()
    corr.evaluate()
    corr.plot(show=True, label="Correlator (bootstrap)")

    print(LogMessage(), "Evaluate covariance")
    corr.evaluate_covmatrix(plot=False)
    corr.corrmat_from_covmat(plot=False)


if __name__ == "__main__":
    main()



import numpy as np
import os
import matplotlib.pyplot as plt
from .utils.rhoUtils import CB_colors, plot_markers, LogMessage
from .transform import combine_base_scalar
from .utils.rhoMath import gauss_fp


def setPlotOpt(plt):
    plt.rcParams["figure.figsize"] = 5, 2
    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "cm"
    plt.rc("xtick", labelsize=22)
    plt.rc("ytick", labelsize=22)
    plt.rcParams.update({"font.size": 22})


def plotwErr(ax, x, y, yerr, label="", markerId=0, colorID=0):
    ax.errorbar(
        x=np.array(x, dtype=float),
        y=np.array(y, dtype=float),
        yerr=np.array(yerr, dtype=float),
        marker=plot_markers[markerId],
        markersize=4.8,
        elinewidth=1.3,
        capsize=2,
        ls="",
        label=label,
        color="black",
        ecolor=CB_colors[colorID],
        markerfacecolor=CB_colors[colorID],
    )


def plotNoErr(ax, x, y, label="", markerId=0, colorID=0):
    ax.errorbar(
        x=np.array(x, dtype=float),
        y=np.array(y, dtype=float),
        marker=plot_markers[markerId],
        markersize=4.8,
        elinewidth=1.3,
        capsize=2,
        ls="",
        label=label,
        color="black",
        ecolor=CB_colors[colorID],
        markerfacecolor=CB_colors[colorID],
    )


def stabilityPlot(invLapW, estar, savePlot=True, plot_live=False):
    setPlotOpt(plt)
    fig, ax = plt.subplots(figsize=(8, 10))
    plt.title(
        r"$E$"
        + "= {:2.2f}  ".format(estar)
        + r" $\;\;\; \sigma$"
        + r" = {:2.2f} ".format(invLapW.par.sigma)
    )
    if invLapW.par.Na > 1:
        _a0label = r"$\alpha = {:1.2f}$".format(invLapW.algorithmPar.alphaA)
    elif invLapW.par.Na == 1:
        _a0label = r"$\rho_\sigma$"

    plotwErr(
        ax,
        invLapW.lambda_list[invLapW.espace_dictionary[estar]],
        invLapW.rho_list[invLapW.espace_dictionary[estar]],
        invLapW.errBoot_list[invLapW.espace_dictionary[estar]],
        markerId=0,
        colorID=0,
        label=_a0label,
    )

    if invLapW.par.Na > 1:
        plotwErr(
            ax,
            invLapW.lambda_list[invLapW.espace_dictionary[estar]],
            invLapW.rho_list_alphaB[invLapW.espace_dictionary[estar]],
            invLapW.errBoot_list_alphaB[invLapW.espace_dictionary[estar]],
            markerId=1,
            colorID=1,
            label=r"$\alpha = {:1.2f}$".format(invLapW.algorithmPar.alphaB),
        )
        if invLapW.par.Na > 2:
            plotwErr(
                ax,
                invLapW.lambda_list[invLapW.espace_dictionary[estar]],
                invLapW.rho_list_alphaC[invLapW.espace_dictionary[estar]],
                invLapW.errBoot_list_alphaC[invLapW.espace_dictionary[estar]],
                markerId=2,
                colorID=2,
                label=r"$\alpha = {:1.2f}$".format(invLapW.algorithmPar.alphaC),
            )

    ax.axhspan(
        ymin=float(
            invLapW.rhoResultHLT[invLapW.espace_dictionary[estar]]
            - invLapW.drho_result[invLapW.espace_dictionary[estar]]
        ),
        ymax=float(
            invLapW.rhoResultHLT[invLapW.espace_dictionary[estar]]
            + invLapW.drho_result[invLapW.espace_dictionary[estar]]
        ),
        alpha=0.3,
        color=CB_colors[4],
    )
    ax.set_xlabel(r"$\lambda$", fontsize=32)
    ax.set_ylabel(r"$\rho_\sigma$", fontsize=32)
    ax.legend()
    ax.set_xscale("log")
    plt.tight_layout()
    if savePlot is True:
        plt.savefig(
            os.path.join(
                invLapW.par.plotpath,
                "LambdaScanE{:2.2e}".format(invLapW.espace_dictionary[estar]) + ".png",
            ),
            dpi=420,
        )
    if plot_live is True:
        plt.show()
    plt.close(fig)
    return


def sharedPlot_stabilityPlusLikelihood(invLapW, estar, savePlot=True, plot_live=False):
    setPlotOpt(plt)
    fig, (ax, ax2) = plt.subplots(
        nrows=2, sharex=True, figsize=(8, 10), gridspec_kw={"height_ratios": [3, 2]}
    )

    plotwErr(
        ax,
        invLapW.lambda_list[invLapW.espace_dictionary[estar]],
        invLapW.rho_list[invLapW.espace_dictionary[estar]],
        invLapW.errBayes_list[invLapW.espace_dictionary[estar]],
        label="Bayesian Error",
        markerId=0,
        colorID=0,
    )

    plotwErr(
        ax,
        invLapW.lambda_list[invLapW.espace_dictionary[estar]],
        invLapW.rho_list[invLapW.espace_dictionary[estar]],
        invLapW.errBoot_list[invLapW.espace_dictionary[estar]],
        label="Bootstrap Error",
        markerId=1,
        colorID=1,
    )

    ax2.set_xlabel(r"$\lambda$", fontsize=32)
    ax.set_ylabel(r"$\rho_\sigma$", fontsize=32)
    ax.legend()
    ax.set_xscale("log")

    plotNoErr(
        ax2,
        invLapW.lambda_list[invLapW.espace_dictionary[estar]],
        invLapW.likelihood_list[invLapW.espace_dictionary[estar]],
        label="NLL",
        markerId=0,
        colorID=0,
    )

    ax2.plot(
        invLapW.lambdaResultBayes[invLapW.espace_dictionary[estar]],
        invLapW.minNLL[invLapW.espace_dictionary[estar]],
        markersize=25,
        marker="*",
        markerfacecolor="red",
        markeredgecolor="black",
        color="red",
        ls="--",
        label="Min NLL",
    )

    ax2.set_ylabel("NLL", fontsize=32)
    plt.subplots_adjust(hspace=0)
    ax.set_title(
        r"$E$"
        + "= {:2.2f}  ".format(estar)
        + r" $\sigma$"
        + " = {:2.2f} ".format(invLapW.par.sigma)
    )
    plt.tight_layout()
    if savePlot is True:
        plt.savefig(
            os.path.join(
                invLapW.par.plotpath,
                "LikelihoodLambdaScanE{:2.2e}".format(invLapW.espace_dictionary[estar])
                + ".png",
            ),
            dpi=420,
        )
    if plot_live is True:
        plt.show()
    plt.close(fig)
    return


def plotLikelihood(invLapW, estar, savePlot=True, plot_live=False):
    setPlotOpt(plt)
    fig, ax = plt.subplots(figsize=(8, 10))
    plt.title(
        r"$E$"
        + "= {:2.2f}  ".format(estar)
        + r" $\;\;\; \sigma$"
        + r" = {:2.2f} ".format(invLapW.par.sigma)
    )
    plotNoErr(
        ax,
        invLapW.lambda_list[invLapW.espace_dictionary[estar]],
        invLapW.likelihood_list[invLapW.espace_dictionary[estar]],
        label="NLL",
        markerId=0,
        colorID=0,
    )
    if invLapW.par.Na > 1:
        plotNoErr(
            ax,
            invLapW.lambda_list[invLapW.espace_dictionary[estar]],
            invLapW.likelihood_list_alphaB[invLapW.espace_dictionary[estar]],
            label="NLL " + r"$\alpha = {:1.2f}$".format(invLapW.algorithmPar.alphaB),
            markerId=1,
            colorID=1,
        )
        if invLapW.par.Na > 2:
            plotNoErr(
                ax,
                invLapW.lambda_list[invLapW.espace_dictionary[estar]],
                invLapW.likelihood_list_alphaC[invLapW.espace_dictionary[estar]],
                label="NLL "
                + r"$\alpha = {:1.2f}$".format(invLapW.algorithmPar.alphaC),
                markerId=2,
                colorID=2,
            )
    ax.plot(
        invLapW.lambdaResultBayes[invLapW.espace_dictionary[estar]],
        invLapW.minNLL[invLapW.espace_dictionary[estar]],
        markersize=25,
        marker="*",
        markerfacecolor="red",
        markeredgecolor="black",
        color="red",
        ls="--",
        label="Minimum",
    )
    ax.set_ylabel("NLL", fontsize=32)
    ax.set_xlabel(r"$\lambda$", fontsize=32)
    ax.legend()
    ax.set_xscale("log")
    plt.tight_layout()
    if savePlot is True:
        plt.savefig(
            os.path.join(
                invLapW.par.plotpath,
                "LikelihoodLambdaOnlyE{:2.2e}".format(invLapW.espace_dictionary[estar])
                + ".png",
            ),
            dpi=420,
        )
    if plot_live is True:
        plt.show()
    plt.close(fig)
    return


def plotAllKernels(invLapW):
    setPlotOpt(plt)
    print(LogMessage(), "Plotting kernel functions")
    _name = "HLTCoefficientsAlpha" + str(float(invLapW.algorithmPar.alphaA)) + ".txt"
    with open(os.path.join(invLapW.par.logpath, _name), "w") as output:
        for _e in range(invLapW.par.Ne):
            print(invLapW.espace[_e], invLapW.gt_HLT[_e], file=output)
            plotKernel(
                invLapW,
                invLapW.gt_HLT[_e],
                ne_=40,
                omega=invLapW.espace[_e],
                label="HLT",
                alpha_=invLapW.algorithmPar.alphaA,
                ker_type=invLapW.par.kerneltype,
            )

    _name = "BayesCoefficientsAlpha" + str(float(invLapW.algorithmPar.alphaA)) + ".txt"
    with open(os.path.join(invLapW.par.logpath, _name), "w") as output:
        for _e in range(invLapW.par.Ne):
            print(invLapW.espace[_e], invLapW.gt_Bayes[_e], file=output)
            plotKernel(
                invLapW,
                invLapW.gt_Bayes[_e],
                ne_=40,
                omega=invLapW.espace[_e],
                label="Bayes",
                alpha_=invLapW.algorithmPar.alphaA,
                ker_type=invLapW.par.kerneltype,
            )


def plotKernel(invLapW, gt_, omega, alpha_, label, ne_=70, ker_type="FULLNORMGAUSS"):
    setPlotOpt(plt)
    fig, ax = plt.subplots(figsize=(8, 10))
    energies = np.linspace(1e-4, invLapW.par.emax * 1.5, ne_)
    kernel = np.zeros(ne_)
    for _e in range(len(energies)):
        kernel[_e] = combine_base_scalar(gt_, invLapW.par, energies[_e])
    plt.plot(
        energies,
        kernel,
        marker="o",
        markersize=3.8,
        ls="--",
        label=label + " Kernel at $\omega$ = " + "{:2.1e}".format(omega),
        color="black",
        markerfacecolor=CB_colors[0],
    )

    y_to_plot = []
    if ker_type == "CAUCHY":

        def ker(k, sigma_, omega_):
            aux = omega_ - k
            aux = aux * aux + sigma_ * sigma_
            aux = sigma_ / aux
            return aux

        y_to_plot = ker(energies, invLapW.par.sigma, omega)

    elif ker_type == "FULLNORMGAUSS":
        y_to_plot = gauss_fp(energies, omega, invLapW.par.sigma, norm="full")
    elif ker_type == "HALFNORMGAUSS":
        y_to_plot = gauss_fp(energies, omega, invLapW.par.sigma, norm="half")

    plt.plot(
        energies,
        y_to_plot,
        ls="-",
        label="Target",
        color="red",
        linewidth=1.2,
    )

    with open(os.path.join(invLapW.par.logpath, f"kernel_{omega}.txt"), "a") as output:
        for e_i in range(len(energies)):
            print(energies[e_i], kernel[e_i], file=output)

    plt.title(
        r" $\sigma$"
        + " = {:2.2f}".format(invLapW.par.sigma)
        + " $\;$ "
        + r"$\alpha$ = {:2.2f}".format(alpha_)
    )
    plt.xlabel(r"$E$")
    plt.legend()
    plt.savefig(
        os.path.join(
            invLapW.par.plotpath,
            label
            + "SmearingKernelSigma{:2.2e}".format(invLapW.par.sigma)
            + "Energy{:2.2e}".format(omega)
            + "Alpha{:2.2f}".format(alpha_)
            + ".png",
        ),
        dpi=400,
    )
    plt.clf()
    return


def plotSpectralDensity(invLapW):
    print(LogMessage(), "Plotting spectral density")
    setPlotOpt(plt)
    fig, ax = plt.subplots(figsize=(8, 10))
    plt.title(r" $\;\;\; \sigma$" + r" = {:2.2f}".format(invLapW.par.sigma))

    plotwErr(
        ax,
        invLapW.espace,
        np.array(invLapW.rhoResultHLT, dtype=float),
        np.array(invLapW.rho_quadrature_err_HLT, dtype=float),
        label="HLT",
        markerId=0,
        colorID=0,
    )

    plotwErr(
        ax,
        invLapW.espace,
        np.array(invLapW.rhoResultBayes, dtype=float),
        np.array(invLapW.rho_quadrature_err_Bayes, dtype=float),
        label="Bayesian",
        markerId=1,
        colorID=1,
    )

    plt.xlabel(r"$E $")
    plt.legend()
    plt.savefig(
        os.path.join(
            invLapW.par.plotpath,
            "RESULT{:2.2e}".format(invLapW.par.sigma) + ".png",
        ),
        dpi=400,
    )
    return



from lsdensities.utils.rhoUtils import LogMessage
from lsdensities.core import integrandSigmaMat
from lsdensities.utils.rhoUtils import Inputs
from mpmath import mp
import time
from scipy.integrate import quad
import numpy as np


def test_time_integration():
    mp.dps = 120
    params = Inputs()
    params.periodicity = "COSH"
    params.kerneltype = "FULLNORMGAUSS"
    params.time_extent = 16

    start = time.time()
    integral = mp.quad(
        lambda x: integrandSigmaMat(x, alpha=0, s=0.1, t1=3, t2=3, E0=0, par=params),
        [0, mp.inf],
        error=False,
        method="tanh-sinh",
    )
    end = time.time()
    print(LogMessage(), float(integral), "in ", end - start, "s")

    integral2, _ = quad(
        lambda x: integrandSigmaMat(x, alpha=0, s=0.1, t1=3, t2=3, E0=0, par=params),
        0.0,
        np.inf,
    )

    print(integral)
    print(integral2)
    assert np.abs(integral2 - integral) < 1e-15



import argparse
from lsdensities.utils.rhoUtils import Inputs


def parse_inputs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-datapath",
        metavar="DataPile",
        type=str,
        help="Path to data file",
        required=True,
    )
    parser.add_argument(
        "--prec",
        metavar="NumericalPrecision",
        type=int,
        help="Numerical precision, approximatively in decimal digits. NOTE: if too high it will be automatically reduced to an optimal value. Default=105",
        default=105,
    )
    parser.add_argument(
        "--tmax",
        metavar="Tmax",
        type=int,
        help="The reconstruction will be performed using correlators c(1), ... c(tmax). If not specified, tmax will be inferred from the time extent of the lattice.",
        default=0,
    )
    parser.add_argument(
        "--outdir", metavar="OutputDirectory", help="Directory for output", default="."
    )
    parser.add_argument(
        "--sigma",
        metavar="GaussianWidth",
        type=float,
        help="Radius of the smearing kernel. Has units of energy.",
        default=0.1,
    )
    parser.add_argument(
        "--nboot",
        metavar="BootstrapSampleSize",
        type=int,
        help="Number of bootstrap samples. Default=300",
        default=300,
    )
    parser.add_argument(
        "--emax",
        type=float,
        help="Maximum energy at which the spectral density is evaluated. Default=1",
        default=1,
    )
    parser.add_argument(
        "--emin",
        type=float,
        help="Maximum energy at which the spectral density is evaluated. Default=1e-2.",
        default=1e-2,
    )
    parser.add_argument(
        "--e0",
        type=float,
        help="Lower integration bound for functional A, Default=0",
        default=0.0,
    )
    parser.add_argument(
        "--Na",
        metavar="Nalpha",
        type=int,
        help="Number of alpha parameters, defining different measure in the functional A, to be used. Allowed values=1,2,3. Default=1, corresponding to alpha=0.",
        default=1,
    )
    parser.add_argument(
        "--ne",
        type=int,
        help="Number of points in energy at which the reconstruction is evaluated, between 0 and emax. Default=20",
        default=20,
    )
    parser.add_argument(
        "--periodicity",
        type=str,
        help="Accepted stirngs are 'EXP' or 'COSH', depending on the correlator being periodic or open.",
        default="EXP",
    )
    parser.add_argument(
        "--A0cut",
        type=float,
        help="Minimum value of A/A0 that is accepted, Default=0.1",
        default=0.1,
    )
    parser.add_argument(
        "--kerneltype",
        type=str,
        help="Accepted stirngs are 'FULLNORMGAUSS', 'HALFNORMGAUSS' or 'CAUCHY', depending on which smearing kernel.",
        default="FULLNORMGAUSS",
    )
    parser.add_argument(
        "--loglevel",
        type=str,
        help="Accepted stirngs are 'WARNING', 'INFO' or 'DEBUG'. Default='WARNING'. Setting 'INFO' leads to an extensive output tracking the details of the scan over lambda and alpha.",
        default="WARNING",
    )
    args = parser.parse_args()
    inputs = Inputs()
    inputs.datapath = args.datapath
    inputs.tmax = args.tmax
    inputs.periodicity = args.periodicity
    inputs.prec = args.prec
    inputs.outdir = args.outdir
    inputs.kerneltype = args.kerneltype
    inputs.num_boot = args.nboot
    inputs.sigma = args.sigma
    inputs.emax = args.emax
    inputs.emin = args.emin
    inputs.e0 = args.e0
    inputs.Ne = args.ne
    inputs.Na = args.Na
    inputs.A0cut = args.A0cut
    inputs.loglevel = args.loglevel
    return inputs


def parse_synthetic_inputs():
    """
    Parse command line for arguments which are turned into attributes of the Inputs class
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--T",
        metavar="TimeExtent",
        type=int,
        help="time extent of the lattice",
        required=True,
    )
    parser.add_argument(
        "--nms",
        metavar="SampleSize",
        type=int,
        help="Number of samples to be generated. Default=100",
        default=100,
    )
    parser.add_argument(
        "--prec",
        metavar="NumericalPrecision",
        type=int,
        help="Numerical precision, approximatively in decimal digits. NOTE: if too high it will be automatically reduced to an optimal value. Default=105",
        default=105,
    )
    parser.add_argument(
        "--tmax",
        metavar="Tmax",
        type=int,
        help="The reconstruction will be performed using correlators c(1), ... c(tmax). If not specified, tmax will be inferred from the time extent of the lattice.",
        default=0,
    )
    parser.add_argument(
        "--outdir", metavar="OutputDirectory", help="Directory for output", default="."
    )
    parser.add_argument(
        "--sigma",
        metavar="GaussianWidth",
        type=float,
        help="Radius of the smearing kernel. Has units of energy.",
        default=0.1,
    )
    parser.add_argument(
        "--nboot",
        metavar="BootstrapSampleSize",
        type=int,
        help="Number of bootstrap samples. Default=300",
        default=300,
    )
    parser.add_argument(
        "--emax",
        type=float,
        help="Maximum energy at which the spectral density is evaluated. Default=1",
        default=1,
    )
    parser.add_argument(
        "--emin",
        type=float,
        help="Maximum energy at which the spectral density is evaluated. Default=1e-2.",
        default=1e-2,
    )
    parser.add_argument(
        "--e0",
        type=float,
        help="Lower integration bound for functional A, Default=0",
        default=0.0,
    )
    parser.add_argument(
        "--Na",
        metavar="Nalpha",
        type=int,
        help="Number of alpha parameters, defining different measure in the functional A, to be used. Allowed values=1,2,3. Default=1, corresponding to alpha=0.",
        default=1,
    )
    parser.add_argument(
        "--ne",
        type=int,
        help="Number of points in energy at which the reconstruction is evaluated, between 0 and emax. Default=20",
        default=20,
    )
    parser.add_argument(
        "--periodicity",
        type=str,
        help="Accepted stirngs are 'EXP' or 'COSH', depending on the correlator being periodic or open.",
        default="EXP",
    )
    parser.add_argument(
        "--A0cut",
        type=float,
        help="Minimum value of A/A0 that is accepted, Default=0.1",
        default=0.1,
    )
    parser.add_argument(
        "--kerneltype",
        type=str,
        help="Accepted stirngs are 'FULLNORMGAUSS', 'HALFNORMGAUSS' or 'CAUCHY', depending on which smearing kernel.",
        default="FULLNORMGAUSS",
    )
    parser.add_argument(
        "--loglevel",
        type=str,
        help="Accepted stirngs are 'WARNING', 'INFO' or 'DEBUG'. Default='WARNING'. Setting 'INFO' leads to an extensive output tracking the details of the scan over lambda and alpha.",
        default="WARNING",
    )
    args = parser.parse_args()
    inputs = Inputs()
    inputs.time_extent = args.T
    inputs.num_samples = args.nms
    inputs.tmax = args.tmax
    inputs.periodicity = args.periodicity
    inputs.prec = args.prec
    inputs.outdir = args.outdir
    inputs.kerneltype = args.kerneltype
    inputs.num_boot = args.nboot
    inputs.sigma = args.sigma
    inputs.emax = args.emax
    inputs.emin = args.emin
    inputs.e0 = args.e0
    inputs.Ne = args.ne
    inputs.Na = args.Na
    inputs.A0cut = args.A0cut
    inputs.loglevel = args.loglevel
    return inputs



from .core import a0_array, hlt_matrix
from .transform import (
    coefficients_ssd,
    get_ssd_averaged_scalar,
    combine_fMf_scalar,
    combine_likelihood,
)
from .abw import gAg
from mpmath import mp, mpf
from .plotutils import (
    stabilityPlot,
    sharedPlot_stabilityPlusLikelihood,
    plotLikelihood,
    plotAllKernels,
    plotSpectralDensity,
)
import numpy as np
from .utils.rhoUtils import Inputs, MatrixBundle, Obs, bcolors, log
from .utils.rhoMath import invert_matrix_ge
import os
import logging


class AlgorithmParameters:
    """
    lambdaMax : Starting value for the scan over lambda.
    lambdaMin : Ending value, unless stopping condition is met.
    lambdaStep : Step in lambda.
    lambdaScanCap : Search for the plateau stops after lambdaScanCap subsequent compatible measurements.
    kfactor : Systematics on value at lambda(reference) are estimated by repeating the calculation at lambda = kfactor lambda(reference).
    resize : if lambda hits zero before hitting lambdaMin, the step is resized. Allows sampling values of lambda at different scales.
    comparisonRatio : Measurements at different lambda are considered compatible if they agree within comparisonRatio * uncertainty (1sigma if comparisonRatio = 1).
    """

    def __init__(
        self,
        alphaA=0,
        alphaB=1 / 2,
        alphaC=1.99,
        lambdaMax=50,
        lambdaStep=25,
        lambdaScanCap=6,
        plateau_id=1,
        kfactor=0.1,
        lambdaMin=1e-6,
        comparisonRatio=0.4,
        resize=4,
    ):
        assert alphaA != alphaB
        assert alphaA != alphaC
        self.alphaA = float(alphaA)
        self.alphaB = float(alphaB)
        self.alphaC = float(alphaC)
        self.lambdaMax = lambdaMax
        self.lambdaStep = lambdaStep
        self.lambdaScanCap = lambdaScanCap
        self.plateau_id = plateau_id
        self.kfactor = kfactor
        # Round trip via a string to avoid introducing spurious precision
        # per recommendations at https://mpmath.org/doc/current/basics.html
        self.alphaAmp = mpf(str(alphaA))
        self.alphaBmp = mpf(str(alphaB))
        self.alphaCmp = mpf(str(alphaC))
        self.lambdaMin = lambdaMin
        self.comparisonRatio = comparisonRatio
        self.resize = resize


class _NormaliseMeasure:
    def __init__(self, par: Inputs, alpha=0, emin=0):
        self.valute_at_E = mp.matrix(par.Ne, 1)
        self.valute_at_E_dictionary = {}  # Auxiliary dictionary: A0espace[n] = A0espace_dictionary[espace[n]] # espace must be float
        self.is_filled = False
        self.alphaMP = alpha
        self.eminMP = emin
        self.par = par

    def evaluate(self, espace_mp):
        log(
            "Computing A0 at all energies with Alpha = {:2.2e}".format(
                float(self.alphaMP)
            ),
        )
        self.valute_at_E = a0_array(espace_mp, self.par, alpha=self.alphaMP)
        for e_id in range(self.par.Ne):
            self.valute_at_E_dictionary[float(espace_mp[e_id])] = self.valute_at_E[e_id]
        self.is_filled = True


class SigmaMatrix:
    def __init__(self, par: Inputs, alphaMP=0):
        self.par = par
        self.tmax = par.tmax
        self.alpha = alphaMP
        self.matrix = mp.matrix(par.tmax, par.tmax)

    def evaluate(self):
        log(" Saving Sigma Matrix ")
        self.matrix = hlt_matrix(
            tmax=self.par.tmax,
            alpha=self.alpha,
            e0=self.par.mpe0,
            type=self.par.periodicity,
            T=self.par.time_extent,
        )


class InverseProblemWrapper:
    def __init__(
        self,
        par: Inputs,
        algorithmPar: AlgorithmParameters,
        matrix_bundle: MatrixBundle,
        correlator: Obs,
        energies,
    ):
        self.par = par
        self.correlator = correlator
        self.algorithmPar = algorithmPar
        self.matrix_bundle = matrix_bundle
        par.Ne = len(energies)
        self.espace = energies
        # Round trip via a string to avoid introducing spurious precision
        # per recommendations at https://mpmath.org/doc/current/basics.html
        self.e0MP = mpf(str(par.e0))
        self.espaceMP = mp.matrix(par.Ne, 1)
        self.sigmaMP = mpf(str(par.sigma))
        self.emaxMP = mpf(str(par.emax))
        self.eminMP = mpf(str(par.emin))
        self.espace_dictionary = {}  #   Usage: espace_dictionary[espace[n]] = n
        self.selectSigmaMat = {}  #   Usage: selectSigmaMat[alpha] = Sigma
        #   Containers for the factor A0
        self.selectA0 = {}
        self.A0_A = _NormaliseMeasure(
            alpha=self.algorithmPar.alphaAmp, emin=self.eminMP, par=self.par
        )
        self.selectA0[algorithmPar.alphaA] = self.A0_A
        if self.par.Na > 1:
            self.A0_B = _NormaliseMeasure(
                alpha=self.algorithmPar.alphaBmp, emin=self.eminMP, par=self.par
            )
            self.selectA0[algorithmPar.alphaB] = self.A0_B
            if self.par.Na > 2:
                self.A0_C = _NormaliseMeasure(
                    alpha=self.algorithmPar.alphaCmp,
                    emin=self.eminMP,
                    par=self.par,
                )
                self.selectA0[algorithmPar.alphaC] = self.A0_C
        self.lambda_list = [[] for _ in range(self.par.Ne)]
        #   First alpha
        self.rho_list = [[] for _ in range(self.par.Ne)]
        self.errBoot_list = [[] for _ in range(self.par.Ne)]
        self.errBayes_list = [[] for _ in range(self.par.Ne)]
        self.gAA0g_list = [[] for _ in range(self.par.Ne)]
        self.likelihood_list = [[] for _ in range(self.par.Ne)]
        self.SigmaMatA = SigmaMatrix(self.par, algorithmPar.alphaA)
        self.selectSigmaMat[algorithmPar.alphaA] = self.SigmaMatA
        #   Second alpha
        if self.par.Na > 1:
            self.rho_list_alphaB = [[] for _ in range(self.par.Ne)]
            self.errBoot_list_alphaB = [[] for _ in range(self.par.Ne)]
            self.errBayes_list_alphaB = [[] for _ in range(self.par.Ne)]
            self.gAA0g_list_alphaB = [[] for _ in range(self.par.Ne)]
            self.likelihood_list_alphaB = [[] for _ in range(self.par.Ne)]
            self.SigmaMatB = SigmaMatrix(self.par, algorithmPar.alphaB)
            self.selectSigmaMat[algorithmPar.alphaB] = self.SigmaMatB
            #   Third alpha
            if self.par.Na > 2:
                self.rho_list_alphaC = [[] for _ in range(self.par.Ne)]
                self.errBoot_list_alphaC = [[] for _ in range(self.par.Ne)]
                self.errBayes_list_alphaC = [[] for _ in range(self.par.Ne)]
                self.gAA0g_list_alphaC = [[] for _ in range(self.par.Ne)]
                self.likelihood_list_alphaC = [[] for _ in range(self.par.Ne)]
                self.SigmaMatC = SigmaMatrix(self.par, algorithmPar.alphaC)
                self.selectSigmaMat[algorithmPar.alphaC] = self.SigmaMatC
        #   Results
        self.minNLL = np.ndarray(
            self.par.Ne, dtype=np.float64
        )  # minimum of Negative Log Likelihood
        self.lambdaResultHLT = np.ndarray(
            self.par.Ne, dtype=np.float64
        )  #   from plateau in lambda
        self.lambdaResultBayes = np.ndarray(
            self.par.Ne, dtype=np.float64
        )  #   from min of NLL
        self.rhoResultHLT = np.ndarray(
            self.par.Ne, dtype=np.float64
        )  #   from plateau in lambda
        self.drho_result = np.ndarray(
            self.par.Ne, dtype=np.float64
        )  #   from plateau in lambda
        self.rhoResultBayes = np.ndarray(
            self.par.Ne, dtype=np.float64
        )  #   from min of NLL
        self.drho_bayes = np.ndarray(self.par.Ne, dtype=np.float64)  #   frin min of NLL
        self.rho_sys_err_HLT = np.ndarray(self.par.Ne, dtype=np.float64)
        self.rho_quadrature_err_HLT = np.ndarray(self.par.Ne, dtype=np.float64)
        self.rho_sys_err_Bayes = np.ndarray(self.par.Ne, dtype=np.float64)
        self.rho_quadrature_err_Bayes = np.ndarray(self.par.Ne, dtype=np.float64)
        self.gt_HLT = [[] for _ in range(self.par.Ne)]  #   from plateau in lambda
        self.gt_Bayes = [[] for _ in range(self.par.Ne)]  #   from min of NLL
        self.aa0 = np.ndarray(
            self.par.Ne, dtype=np.float64
        )  # A / A0 for the result (HLT only)
        #   Control variables
        self.espace_is_filled = False
        self.A0_is_filled = False
        self.result_is_filled = np.full(par.Ne, False, dtype=bool)
        return
        # - - - - - - - - - - - - - - - End of INIT - - - - - - - - - - - - - - - #

    def fillEspaceMP(self):
        """
        Fill array of energies : array[int] = energy
        and provides a dictionary such that dictionary[array[int]] = int
        """
        for e_id in range(self.par.Ne):
            self.espaceMP[e_id] = mpf(str(self.espace[e_id]))
            self.espace_dictionary[self.espace[e_id]] = e_id
        self.espace_is_filled = True
        return

    def prepareHLT(self):
        """
        Creates directory for output
        Fills array of energies and corresponding dictionary
        Evaluates A0 at each energy
        Evaluates the matrix Sigma
        """
        self.fillEspaceMP()
        self.A0_A.evaluate(self.espaceMP)
        self.SigmaMatA.evaluate()
        with open(
            os.path.join(self.par.logpath, "InverseProblemLOG_AlphaA.log"), "w"
        ) as output:
            print(
                "# estar\t",
                "lambda\t",
                "rho\t",
                "errBayes\t",
                "errBoot\t",
                "A/A0\t",
                "likelihood\n",
                file=output,
            )
        if self.par.Na > 1:
            self.A0_B.evaluate(self.espaceMP)
            self.SigmaMatB.evaluate()
            with open(
                os.path.join(self.par.logpath, "InverseProblemLOG_AlphaB.log"), "w"
            ) as output:
                print(
                    "# estar\t",
                    "lambda\t",
                    "rho\t",
                    "errBayes\t",
                    "errBoot\t",
                    "A/A0\t",
                    "likelihood\n",
                    file=output,
                )
            if self.par.Na > 2:
                self.A0_C.evaluate(self.espaceMP)
                self.SigmaMatC.evaluate()
                with open(
                    os.path.join(self.par.logpath, "InverseProblemLOG_AlphaC.log"), "w"
                ) as output:
                    print(
                        "# estar\t",
                        "lambda\t",
                        "rho\t",
                        "errBayes\t",
                        "errBoot\t",
                        "A/A0\t",
                        "likelihood\n",
                        file=output,
                    )
        self.A0_is_filled = True
        return

    def _store(
        self,
        estar,
        rho,
        errBayes,
        errBoot,
        likelihood,
        gag,
        lambda_,
        whichAlpha="A",
    ):
        """
        Each call appends a line to a file named InverseProblemLog_{alpha} containing
        estar, rho, errBayes, errBoot, likelihood, g(A/A0)g, lambda
        """
        if whichAlpha == "A":
            self.rho_list[self.espace_dictionary[estar]].append(rho)
            self.errBayes_list[self.espace_dictionary[estar]].append(errBayes)
            self.errBoot_list[self.espace_dictionary[estar]].append(errBoot)
            self.gAA0g_list[self.espace_dictionary[estar]].append(
                gag
                / self.selectA0[self.algorithmPar.alphaA].valute_at_E_dictionary[estar]
            )
            self.likelihood_list[self.espace_dictionary[estar]].append(likelihood)
            with open(
                os.path.join(self.par.logpath, "InverseProblemLOG_AlphaA.log"), "a"
            ) as output:
                print(
                    float(estar),
                    float(lambda_),
                    float(rho),
                    float(errBayes),
                    float(errBoot),
                    float(gag),
                    float(likelihood),
                    file=output,
                )
        if whichAlpha == "B":
            self.rho_list_alphaB[self.espace_dictionary[estar]].append(rho)
            self.errBayes_list_alphaB[self.espace_dictionary[estar]].append(errBayes)
            self.errBoot_list_alphaB[self.espace_dictionary[estar]].append(errBoot)
            self.gAA0g_list_alphaB[self.espace_dictionary[estar]].append(
                gag
                / self.selectA0[self.algorithmPar.alphaA].valute_at_E_dictionary[estar]
            )
            self.likelihood_list_alphaB[self.espace_dictionary[estar]].append(
                likelihood
            )
            with open(
                os.path.join(self.par.logpath, "InverseProblemLOG_AlphaB.log"), "a"
            ) as output:
                print(
                    float(estar),
                    float(lambda_),
                    float(rho),
                    float(errBayes),
                    float(errBoot),
                    float(gag),
                    float(likelihood),
                    file=output,
                )
        if whichAlpha == "C":
            self.rho_list_alphaC[self.espace_dictionary[estar]].append(rho)
            self.errBayes_list_alphaC[self.espace_dictionary[estar]].append(errBayes)
            self.errBoot_list_alphaC[self.espace_dictionary[estar]].append(errBoot)
            self.gAA0g_list_alphaC[self.espace_dictionary[estar]].append(
                gag
                / self.selectA0[self.algorithmPar.alphaA].valute_at_E_dictionary[estar]
            )
            self.likelihood_list_alphaC[self.espace_dictionary[estar]].append(
                likelihood
            )
            with open(
                os.path.join(self.par.logpath, "InverseProblemLOG_AlphaC.log"), "a"
            ) as output:
                print(
                    float(estar),
                    float(lambda_),
                    float(rho),
                    float(errBayes),
                    float(errBoot),
                    float(gag),
                    float(likelihood),
                    file=output,
                )
        return

    def _areRangesCompatible(self, x, deltax, y, deltay):
        """
        Checks whether ranges are compatible.
        """
        x2 = x + deltax
        x1 = x - deltax
        y1 = y - deltay
        y2 = y + deltay
        return x1 <= y2 and y1 <= x2

    def _flagResult(self, *args):
        return args

        # - - - - - - - - - - - - - - - Main function: given lambda computes rho_s - - - - - - - - - - - - - - - #

    def lambdaToRho(self, lambda_, estar_, alpha_):
        """
        For a given lambda, at each energy and value of alpha
        computes and returns the following
            rho_estar : the result for the smeared spectral density
            drho_estar_Bayes : Bayesian error
            drho_estar_Bootstrap : frequentist error
            likelihood_estar : likelihood of the result
            gAg_estar : the scalar product gAg
            _g_t_estar : the vector of coefficients giving the result

        The coefficients are obtained by inverting the matrix
        S + factor B
        where factor = lambda A0 / Bnorm.
        Bnorm makes B dimensionless.
        """
        import time

        _Bnorm = self.matrix_bundle.bnorm / (estar_ * estar_)
        _factor = (
            lambda_ * self.selectA0[float(alpha_)].valute_at_E_dictionary[estar_]
        ) / _Bnorm
        log("Normalising factor A*l/B = {:2.2e}".format(float(_factor)))

        S = self.selectSigmaMat[float(alpha_)].matrix
        _Matrix = S + (_factor * self.matrix_bundle.B)
        start_time = time.time()
        _MatrixInv = invert_matrix_ge(_Matrix)
        end_time = time.time()
        log(
            "Time ::: Matrix inverted in {:4.4f}".format(end_time - start_time),
            "s",
        )
        start_time = time.time()
        _g_t_estar = coefficients_ssd(_MatrixInv, self.par, estar_, alpha=alpha_)
        end_time = time.time()
        log(
            "Time ::: Coefficients computed in {:4.4f}".format(end_time - start_time),
            "s",
        )
        rho_estar, drho_estar_Bootstrap = get_ssd_averaged_scalar(
            _g_t_estar, self.correlator.mpsample, self.par
        )
        start_time = time.time()
        log(
            "Time ::: Bootstrapped result in {:4.4f}".format(start_time - end_time),
            "s",
        )
        gAg_estar = gAg(S, _g_t_estar, estar_, alpha_, self.par)

        varianceRho = combine_fMf_scalar(
            gt=_g_t_estar, params=self.par, estar=estar_, alpha=alpha_
        )
        log("\t\t gt ft = ", float(varianceRho))
        log(
            "\t\t A0 is ",
            float(self.selectA0[float(alpha_)].valute_at_E_dictionary[estar_]),
        )
        varianceRho = mp.fsub(
            float(self.selectA0[float(alpha_)].valute_at_E_dictionary[estar_]),
            varianceRho,
        )
        log("\t\t A0 - gt ft = {:2.2e}".format(float(varianceRho)))
        varianceRho = mp.fdiv(varianceRho, _factor)
        varianceRho = mp.fdiv(varianceRho, mpf(2))
        drho_estar_Bayes = mp.sqrt(abs(varianceRho))

        log(
            "\t \t lambdaToRho ::: Central Value = {:2.4e}".format(float(rho_estar)),
        )
        log(
            "\t \t lambdaToRho ::: Bayesian Error = {:2.4e}".format(
                float(drho_estar_Bayes)
            ),
        )
        log(
            "\t \t lambdaToRho ::: Bootstrap Error   = {:2.4e}".format(
                float(drho_estar_Bootstrap)
            ),
        )

        fullCov_inv = _MatrixInv * _factor

        #   Compute the likelihood
        likelihood_estar = combine_likelihood(
            fullCov_inv, self.par, self.correlator.mpcentral
        )
        likelihood_estar *= 0.5
        det = mp.det(_Matrix / _factor)
        likelihood_estar = mp.fadd(likelihood_estar, 0.5 * mp.log(det))
        likelihood_estar = mp.fadd(
            likelihood_estar, (self.par.tmax * mp.log(2 * mp.pi)) * 0.5
        )
        log(
            "\t \t lambdaToRho ::: NLL = {:2.4e}".format(float(likelihood_estar)),
        )

        return (
            rho_estar,
            drho_estar_Bayes,
            drho_estar_Bootstrap,
            likelihood_estar,
            gAg_estar,
            _g_t_estar,
        )

    # - - - - - - - - - - - - - - - Scan over parameters: Lambda, Alpha (optional) - - - - - - - - - - - - - - - #

    def scanParameters(self, estar_):
        """
        This function will scan over lambda and, if specified, alpha, until conditions are met
        The stopping conditions are one of the following:
            - compatibility is achieved for a subsequent number of values of lambda specified by self.algorithmPar.lambdaScanCap : this is the intended way.
            OR
            - Reaching self.algorithmPar.lambdaMin : if compatibility conditions were never met. Raises a Warning.
        The compatibility between results at different lambda (or alpha) is achieved when the following conditions are simultaneously met:
            - A/A0 < self.par.A0cut
            AND
            - rho(lambda) = rho(lambda') within N sigma, where N is given by self.par.comparisonRatio.
              Setting comparisonRatio=1 means results are considered compatible when their errorbands overlap.
              Using values smaller than 1 can be useful since rho at different values of lambda can be very correlated. The default value
              is 0.3 which has been set empirically.
            AND
            - rho(lambda , alpha) = rho(lambda, alpha') = rho(lambda, alpha'') within 1 sigma, if more values of alpha are used.
        The scan is done between lambdaMax and lambdaMin. The initial step is very large, but it is resized whenever lambda becomes negative
        to allow a fast but meaningful scan than a fixed step could achieve.
        """
        lambda_ = self.algorithmPar.lambdaMax
        lambda_step = self.algorithmPar.lambdaStep
        _cap = self.algorithmPar.lambdaScanCap
        _resize = self.algorithmPar.resize
        _countPositiveResult = 0
        _compRatio = self.algorithmPar.comparisonRatio
        _plateau_id = self.algorithmPar.plateau_id

        log(" --- ")
        log("At Energy {:2.2e}".format(estar_))
        log(
            "Setting Lambda ::: Lambda (0,inf) = {:1.3e}".format(float(lambda_)),
        )
        log(
            "Setting Lambda ::: Lambda (0,1) = {:1.3e}".format(
                float(lambda_ / (1 + lambda_))
            ),
        )

        #   First runs on initial values, then loops

        self.lambda_list[self.espace_dictionary[estar_]].append(lambda_)

        _rho, _errBayes, _errBoot, _likelihood, _gAg, _gt = self.lambdaToRho(
            lambda_, estar_, self.algorithmPar.alphaAmp
        )
        self._store(
            estar_,
            _rho,
            _errBayes,
            _errBoot,
            _likelihood,
            _gAg,
            lambda_,
            whichAlpha="A",
        )
        if self.par.Na > 1:
            _rhoB, _errBayesB, _errBootB, _likelihoodB, _gAgB, _gtB = self.lambdaToRho(
                lambda_, estar_, self.algorithmPar.alphaBmp
            )
            self._store(
                estar_,
                _rhoB,
                _errBayesB,
                _errBootB,
                _likelihoodB,
                _gAgB,
                lambda_,
                whichAlpha="B",
            )
            if self.par.Na > 2:
                (
                    _rhoC,
                    _errBayesC,
                    _errBootC,
                    _likelihoodC,
                    _gAgC,
                    _gtC,
                ) = self.lambdaToRho(lambda_, estar_, self.algorithmPar.alphaCmp)
                self._store(
                    estar_,
                    _rhoC,
                    _errBayesC,
                    _errBootC,
                    _likelihoodC,
                    _gAgC,
                    lambda_,
                    whichAlpha="C",
                )

        #   Flag these until better results (Bayesian)
        minNLL, lambdaStarBayes, rhoBayes, drhoBayes, gtBAYES = self._flagResult(
            _likelihood, lambda_, _rho, _errBayes, _gt
        )

        #   #   #   #   #   #   #   #   Loops over values of lambda #   #   #   #   #   #   #   #
        lambda_ -= lambda_step
        while _countPositiveResult < _cap and lambda_ > self.algorithmPar.lambdaMin:
            #   -   -   -   -   -   -
            log(
                "Setting Lambda ::: Lambda (0,inf) = {:1.3e}".format(float(lambda_)),
            )
            log(
                "Setting Lambda ::: Lambda (0,1) = {:1.3e}".format(
                    float(lambda_ / (1 + lambda_))
                ),
            )
            self.lambda_list[self.espace_dictionary[estar_]].append(lambda_)

            #   Flag these until better results (HLT). Inside the loop contrary to the Bayesian equivalent
            #   because _countPositiveResult can be set to zero inside the loop
            if _countPositiveResult == 0:
                lambdaStarHLT, rhoHLT, drhoHLT, gtHLT, gag_flag = self._flagResult(
                    lambda_, _rho, _errBoot, _gt, _gAg
                )
            #   -   -   -   -   -   -

            log(
                "\t Setting Alpha ::: First Alpha = ",
                self.algorithmPar.alphaA,
            )
            (
                _rhoUpdated,
                _errBayesUpdated,
                _errBootUpdated,
                _likelihoodUpdated,
                _gAgUpdated,
                _gtUpdated,
            ) = self.lambdaToRho(lambda_, estar_, self.algorithmPar.alphaAmp)
            self._store(
                estar_,
                _rhoUpdated,
                _errBayesUpdated,
                _errBootUpdated,
                _likelihoodUpdated,
                _gAgUpdated,
                lambda_,
                whichAlpha="A",
            )
            if self.par.Na > 1:
                log(
                    "\t Setting Alpha ::: Second Alpha = ",
                    self.algorithmPar.alphaB,
                )
                (
                    _rhoUpdatedB,
                    _errBayesUpdatedB,
                    _errBootUpdatedB,
                    _likelihoodUpdatedB,
                    _gAgUpdatedB,
                    _gtUpdatedB,
                ) = self.lambdaToRho(lambda_, estar_, self.algorithmPar.alphaBmp)
                self._store(
                    estar_,
                    _rhoUpdatedB,
                    _errBayesUpdatedB,
                    _errBootUpdatedB,
                    _likelihoodUpdatedB,
                    _gAgUpdatedB,
                    lambda_,
                    whichAlpha="B",
                )
                _AB_Overlap = self._areRangesCompatible(
                    _rhoUpdated, _errBootUpdated, _rhoUpdatedB, _errBootUpdatedB
                )
                if self.par.Na > 2:
                    log(
                        "\t Setting Alpha ::: Third Alpha = ",
                        self.algorithmPar.alphaC,
                    )
                    (
                        _rhoUpdatedC,
                        _errBayesUpdatedC,
                        _errBootUpdatedC,
                        _likelihoodUpdatedC,
                        _gAgUpdatedC,
                        _gtUpdatedC,
                    ) = self.lambdaToRho(lambda_, estar_, self.algorithmPar.alphaCmp)
                    self._store(
                        estar_,
                        _rhoUpdatedC,
                        _errBayesUpdatedC,
                        _errBootUpdatedC,
                        _likelihoodUpdatedC,
                        _gAgUpdatedC,
                        lambda_,
                        whichAlpha="C",
                    )
                    _AC_Overlap = self._areRangesCompatible(
                        _rhoUpdated, _errBootUpdated, _rhoUpdatedC, _errBootUpdatedC
                    )

            newLambda_Overlap = self._areRangesCompatible(
                _rhoUpdated, _compRatio * _errBootUpdated, _rho, _compRatio * _errBoot
            )  # comparison with previous lambda

            flagLambda_Overlap = self._areRangesCompatible(
                _rhoUpdated, _compRatio * _errBootUpdated, rhoHLT, _compRatio * drhoHLT
            )  # comparison with flagged lambda

            if self.par.Na > 1:
                newLambda_Overlap_B = self._areRangesCompatible(
                    _rhoUpdatedB,
                    _compRatio * _errBootUpdatedB,
                    _rhoB,
                    _compRatio * _errBootB,
                )  # comparison with previous lambda

                flagLambda_Overlap_B = self._areRangesCompatible(
                    _rhoUpdatedB,
                    _compRatio * _errBootUpdatedB,
                    rhoHLT,
                    _compRatio * drhoHLT,
                )  # comparison with flagged lambda

                if self.par.Na > 2:
                    newLambda_Overlap_C = self._areRangesCompatible(
                        _rhoUpdatedC,
                        _compRatio * _errBootUpdatedC,
                        _rhoC,
                        _compRatio * _errBootC,
                    )  # comparison with previous lambda

                    flagLambda_Overlap_C = self._areRangesCompatible(
                        _rhoUpdatedC,
                        _compRatio * _errBootUpdatedC,
                        rhoHLT,
                        _compRatio * drhoHLT,
                    )  # comparison with flagged lambda

            if (
                _likelihoodUpdated < minNLL
            ):  # NLL, if less than before; flag the results
                (
                    minNLL,
                    lambdaStarBayes,
                    rhoBayes,
                    drhoBayes,
                    gtBAYES,
                ) = self._flagResult(
                    _likelihoodUpdated,
                    lambda_,
                    _rhoUpdated,
                    _errBayesUpdated,
                    _gtUpdated,
                )

            _skip = False
            #   Checks compatibility between different alphas
            if self.par.Na > 1 and not _AB_Overlap:
                log("\t First and Second Alpha not compatible")
                _skip = True
                if newLambda_Overlap_B is False:
                    log(
                        "\t Result at Second Alpha not compatible with previous lambda at Second Alpha"
                    )
                    _skip = True
            if self.par.Na > 2 and not _AC_Overlap:
                log("\t First and Third Alpha not compatible")
                _skip = True
                if newLambda_Overlap_C is False:
                    log(
                        "\t Result at Third Alpha not compatible with previous lambda at Third Alpha"
                    )
                    _skip = True

            #   Checks if Rho at this lambda overlaps with Rho at flagged lambda
            if newLambda_Overlap is False:
                log(
                    "\t Result at this Lambda does not overlap with previous: REJECTING result",
                )
                _skip = True
            if flagLambda_Overlap is False:
                log(
                    "\t Result at this Lambda does not overlap with flagged: REJECTING result",
                )
                _skip = True
                if self.par.Na > 1 and not flagLambda_Overlap_B:
                    log(
                        "\t Result at this Lambda, Second Alpha, does not overlap with flagged: REJECTING result",
                    )
                    _skip = True
                    if self.par.Na > 2 and not flagLambda_Overlap_C:
                        log(
                            "\t Result at this Lambda, Third Alpha, does not overlap with flagged: REJECTING result",
                        )
                        _skip = True

            #   Checks A/A0 is acceptable
            if (
                _gAgUpdated
                / self.selectA0[self.algorithmPar.alphaA].valute_at_E_dictionary[estar_]
                > self.par.A0cut
            ):
                log(
                    "\t A/A0 is too large: rejecting result  (",
                    float(
                        _gAgUpdated
                        / self.selectA0[
                            self.algorithmPar.alphaA
                        ].valute_at_E_dictionary[estar_]
                    ),
                    ")",
                )
                _skip = True

            #   Having analysed all possible stopping conditions, proceed

            if _skip is False:
                #   Flag the first compatible result, because it is the one with the smaller error
                if (
                    _countPositiveResult == self.algorithmPar.plateau_id
                ):  #   At future alphas we compare with rho_s at _countPositiveResult = 1. This can be changed.
                    (
                        lambdaStarHLT,
                        rhoHLT,
                        drhoHLT,
                        gtHLT,
                        gag_flag,
                    ) = self._flagResult(
                        lambda_, _rho, _errBoot, _gtUpdated, _gAgUpdated
                    )
                _countPositiveResult += 1
                log(
                    f"{bcolors.OKGREEN}Stopping Condition{bcolors.ENDC}",
                    _countPositiveResult,
                    "/",
                    _cap,
                )
            else:
                _countPositiveResult = 0

            #   Update variables before restarting the loop
            _rho = _rhoUpdated
            _errBoot = _errBootUpdated
            lambda_ -= lambda_step
            #   Resize lambda_step
            if lambda_ <= 0:
                lambda_step /= _resize
                lambda_ += lambda_step * (_resize - 1 / _resize)
                log(
                    "Resize LambdaStep to ",
                    lambda_step,
                    "Setting Lambda = ",
                    lambda_,
                )

            if lambda_ < self.algorithmPar.lambdaMin:
                logging.warning(
                    f"{bcolors.WARNING}Warning{bcolors.ENDC} ::: Stopping ::: Reached lower limit for lambda. Try decreasing 'algorithmPar.lambdaMin' or increase the smearing radius.",
                )

        #   End of WHILE
        if _countPositiveResult == 0:
            logging.warning(
                f"{bcolors.WARNING}WARNING{bcolors.ENDC} ::: Could NOT find a plateau in lambda",
            )

        #   hlt
        self.lambdaResultHLT[self.espace_dictionary[estar_]] = lambdaStarHLT
        self.rhoResultHLT[self.espace_dictionary[estar_]] = rhoHLT
        self.drho_result[self.espace_dictionary[estar_]] = drhoHLT
        self.gt_HLT[self.espace_dictionary[estar_]] = gtHLT
        self.aa0[self.espace_dictionary[estar_]] = gag_flag
        #   bayesian
        self.minNLL[self.espace_dictionary[estar_]] = minNLL
        self.lambdaResultBayes[self.espace_dictionary[estar_]] = lambdaStarBayes
        self.rhoResultBayes[self.espace_dictionary[estar_]] = rhoBayes
        self.drho_bayes[self.espace_dictionary[estar_]] = drhoBayes
        self.gt_Bayes[self.espace_dictionary[estar_]] = gtBAYES

        return (
            lambdaStarHLT,
            rhoHLT,
            drhoHLT,
            minNLL,
            lambdaStarBayes,
            rhoBayes,
            drhoBayes,
            gtHLT,
            gtBAYES,
            gag_flag,
        )

    def estimate_sys_error(self, e_i):
        _this_y_HLT = self.rhoResultHLT[e_i]  # rho at lambda*
        _that_y_HLT, _, _, _, _, _ = self.lambdaToRho(
            self.lambdaResultHLT[e_i] * self.algorithmPar.kfactor,
            self.espace[e_i],
            alpha_=0,
        )

        _this_y_Bayes = self.rhoResultBayes[e_i]  # rho at lambda*
        _that_y_Bayes, _, _, _, _, _ = self.lambdaToRho(
            self.lambdaResultBayes[e_i] * self.algorithmPar.kfactor,
            self.espace[e_i],
            alpha_=0,
        )

        self.rho_sys_err_HLT[e_i] = abs(_this_y_HLT - _that_y_HLT) / 2
        self.rho_quadrature_err_HLT[e_i] = np.sqrt(
            self.rho_sys_err_HLT[e_i] ** 2 + self.drho_result[e_i] ** 2
        )

        self.rho_sys_err_Bayes[e_i] = abs(_this_y_Bayes - _that_y_Bayes) / 2
        self.rho_quadrature_err_Bayes[e_i] = np.sqrt(
            self.rho_sys_err_Bayes[e_i] ** 2 + self.drho_bayes[e_i] ** 2
        )

        return self.rho_sys_err_HLT[e_i], self.rho_sys_err_Bayes[e_i]

    def run(self, savePlots=True, livePlots=False):
        with open(os.path.join(self.par.logpath, "ResultHLT.txt"), "w") as output:
            print(
                "# Energy \t Lambda(HLT) \t Rho(HLT) \t Stat(HLT) \t Sys(HLT) \t Quadrature \t A/A0",
                file=output,
            )
        with open(os.path.join(self.par.logpath, "ResultBayes.txt"), "w") as output:
            print(
                "# Energy \t Lambda(Bayes) \t Rho(Bayes) \t Stat(Bayes) \t Sys(Bayes) \t Quadrature \t NLL",
                file=output,
            )

        for e_i in range(self.par.Ne):
            self.scanParameters(self.espace[e_i])
            self.estimate_sys_error(e_i)
            with open(os.path.join(self.par.logpath, "ResultHLT.txt"), "a") as output:
                print(
                    self.espace[e_i],
                    self.lambdaResultHLT[e_i],
                    float(self.rhoResultHLT[e_i]),
                    float(self.drho_result[e_i]),
                    float(self.rho_sys_err_HLT[e_i]),
                    float(self.rho_quadrature_err_HLT[e_i]),
                    float(self.aa0[e_i]),
                    file=output,
                )
            with open(os.path.join(self.par.logpath, "ResultBayes.txt"), "a") as output:
                print(
                    self.espace[e_i],
                    self.lambdaResultBayes[e_i],
                    float(self.rhoResultBayes[e_i]),
                    float(self.drho_bayes[e_i]),
                    float(self.rho_sys_err_Bayes[e_i]),
                    float(self.rho_quadrature_err_Bayes[e_i]),
                    float(self.minNLL[e_i]),
                    file=output,
                )

        return 0

    def stabilityPlot(
        self,
        generateHLTscan=True,
        generateLikelihoodShared=True,
        generateLikelihoodPlot=True,
        generateKernelsPlot=True,
    ):
        for e_i in range(self.par.Ne):
            if generateHLTscan is True:
                stabilityPlot(self, self.espace[e_i], savePlot=True, plot_live=False)
            if generateLikelihoodShared is True:
                sharedPlot_stabilityPlusLikelihood(
                    self, self.espace[e_i], savePlot=True, plot_live=False
                )
            if generateLikelihoodPlot is True:
                plotLikelihood(self, self.espace[e_i], savePlot=True, plot_live=False)
            if generateKernelsPlot is True:
                plotAllKernels(self)

    def plotResult(self):
        plotSpectralDensity(self)
        return



import numpy as np
import matplotlib.pyplot as plt
import random as rd
import os
import time
from mpmath import mp, mpf
import hashlib
import logging

target_result_precision = 1e-8

#   #   #   #   #   #  ----- logger -----   #   #   #   #   #   #

logger = logging.getLogger("log")
logger.setLevel(logging.WARNING)  # default log level is WARNING
logger.propagate = False  # avoid duplication of messages in some cases

if not logger.hasHandlers():
    stream_handler = logging.StreamHandler()

    # custom formatter class to include elapsed time
    class CustomFormatter(logging.Formatter):
        def __init__(self, *args, **kwargs):
            self.start_time = time.time()
            super().__init__(*args, **kwargs)

        def format(self, record):
            elapsed_time_ms = time.time() - self.start_time
            record.elapsed_time = "{:.3f} s".format(elapsed_time_ms)
            return super().format(record)

    # set custom formatter for the handler
    formatter = CustomFormatter("Message ::: %(elapsed_time)s - %(message)s")
    stream_handler.setFormatter(formatter)

    # add the handler to the logger
    logger.addHandler(stream_handler)


def log(*args, **kwargs):
    level = kwargs.pop("level", logging.INFO)
    msg = " ".join(map(str, args))
    logger.log(level, msg, **kwargs)


start_time = time.time()


def LogMessage():
    return "Message ::: {:2.5f}".format(time.time() - start_time) + " s :::"


def end():
    print(LogMessage(), "Exit")
    exit()


#   #   #   #   #   #   #   #   #


def generate_seed(par):
    """
    :param par: Input class instance
    Generates seed from a hash of the inputs
    """
    # Concatenate the input parameters into a string
    input_string = f"{par.emin}{par.emax}{par.Ne}{par.time_extent}{par.sigma}"

    # Encode the string to bytes
    encoded_string = input_string.encode("utf-8")

    # Calculate the SHA-256 hash
    sha256_hash = hashlib.sha256(encoded_string).hexdigest()

    return sha256_hash


def create_out_paths(par):
    if not os.path.exists(par.outdir):
        os.mkdir(par.outdir)
    dir = os.path.join(par.outdir, par.directoryName)
    if not os.path.exists(dir):
        os.mkdir(dir)
    plotpath = os.path.join(dir, "Plots")
    logpath = os.path.join(dir, "Logs")
    if not os.path.exists(plotpath):
        os.mkdir(plotpath)
    if not os.path.exists(logpath):
        os.mkdir(logpath)
    return plotpath, logpath


def ranvec(vec, dim, a, b):
    for j in range(0, dim):
        vec[j] = rd.randint(a, b - 1)
    return vec


class Obs:
    """
    Class for an array of observables
    T: lenght of the array.
    tmax: highest element of the array that is used for analysis.
    is_resampled: True if the samples are already an average (e.g. jackknife, bootstrap), False otherwise.

    Attributes:
    central: values of the array.
    err: error on central.
    sigma: std, which corresponds to err if is_resampled = True. When if_resampled = False, the
           two are related by sqrt(measurements).
    nms: number of measurements,
    sample: an array of measurements (len = nms) for each array in the observable.
    cov: covariance matrix
    corrmat: correlation matrix
    mpsample: sample, converted into mp variables, and of leght reduced from T to tmax
    mpcov: cov from mpsample
    mpcentral: central from mpsample
    """

    def __init__(self, T: int, tmax: int, nms: int = 1, is_resampled=False):
        self.central = np.zeros(T)  # Central value of the sample
        self.err = np.zeros(T)  # Error on the central value
        self.sigma = np.zeros(T)  # Variance of the sample
        self.T = T  # number of time slices
        self.tmax = tmax  # Max t we use
        self.nms = nms
        self.sample = np.zeros((nms, T))  # Sample elements
        self.cov = np.zeros((T, T))  # Cov matrix estimated from sample
        self.corrmat = np.zeros((T, T))  # Corr matrix estimated from sample
        self.is_resampled = is_resampled
        self.mpsample = mp.matrix(self.nms, self.tmax)
        self.mpcov = mp.matrix(self.tmax, self.tmax)
        self.mpcentral = mp.matrix(self.tmax, 1)

    def evaluate(self):
        """
        From sample, computes central and err and store them into
        self.central, self.err
        """
        for i in range(self.T):
            self.central[i], self.sigma[i] = (
                np.average(self.sample[:, i]),
                np.std(self.sample[:, i], ddof=1),
            )
        if self.is_resampled is False:
            self.err = self.sigma / np.sqrt(self.nms)
        if self.is_resampled is True:
            self.err = self.sigma

    def evaluate_covmatrix(self, plot=False):
        """
        From sample, computes covariance matrix in self.cov
        """
        sample_matrix = np.array(self.sample).T
        self.cov = np.cov(sample_matrix, bias=False)
        if plot:
            plt.imshow(self.cov, cmap="viridis")
            plt.colorbar()
            plt.show()
        return self.cov

    def corrmat_from_covmat(self, plot=False):
        """
        Computes the correlation matrix from the covariance matrix
        and saves it into self.corrmat
        """
        for vi in range(self.T):
            for vj in range(self.T):
                self.corrmat[vi][vj] = self.cov[vi][vj] / (
                    self.sigma[vi] * self.sigma[vj]
                )
        if plot is True:
            plt.imshow(self.corrmat)
            plt.colorbar()
            plt.show()

    def fill_mp_sample(self):
        """
        This operation also includes the shifting of the correlator index
        so that corr(0) is never used
        """
        for n in range(self.nms):
            for i in range(self.tmax):  # tmax = T/2 if folded otherwise T-1
                self.mpsample[n, i] = mpf(str(self.sample[n][i + 1]))
        #   Get cov for B matrix
        self.mpcov = mp.matrix(self.tmax)
        for i in range(self.tmax):
            self.mpcentral[i] = self.central[i + 1]
            for j in range(self.tmax):
                self.mpcov[i, j] = mpf(str(self.cov[i + 1][j + 1]))

    def fill_mp_sample_NOSHIFT(self):
        for n in range(self.nms):
            for i in range(self.tmax):  # tmax = T/2 if folded otherwise T-1
                self.mpsample[n, i] = mpf(str(self.sample[n][i]))
        #   Get cov for B matrix
        self.mpcov = mp.matrix(self.tmax)
        for i in range(self.tmax):
            self.mpcentral[i] = self.central[i]
            for j in range(self.tmax):
                self.mpcov[i, j] = mpf(str(self.cov[i][j]))

    def plot(self, show=True, logscale=True, label=None, yscale=1):
        plt.grid(alpha=0.1)
        if logscale is True:
            plt.yscale("log")
        plt.errorbar(
            x=list(range(0, self.T)),
            y=self.central / yscale,
            yerr=self.err / yscale,
            marker="o",
            markersize=1.5,
            elinewidth=1,
            ls="",
            label=label,
            color="b",
        )
        plt.tight_layout()
        if label is not None:
            plt.legend()
        if show is True:
            plt.show()


def print_hlt_format(mtobs, T, nms, filename, directory):
    cout = os.path.join(directory, filename)
    with open(cout, "w") as output:
        print(nms, T, T, "2", "3", file=output)
        for j in range(0, nms):
            for i in range(0, T):
                print(i, mtobs[j, i], file=output)


def read_datafile(par, resampled=False):  # (filename_, directory_):
    """
    The input file has a header with time_extent and number of measurements.
    then data config by config. Example:
    #   32  100
    #   0   corr[0]
    #   1   corr[1]
    #   ...
    #   31  corr[31]
    #   0   corr[0]
    #   ... so on
    """
    datapath = par.datapath
    with open(datapath, "r") as file:
        header = next(file).strip()
        print(LogMessage(), "Reading file :::", "Header: ", header)
        header_nms = int(header.split(" ")[0])
        header_T = int(header.split(" ")[1])
        print(LogMessage(), "Reading file :::", "Time extent ", header_T)
        print(LogMessage(), "Reading file :::", "Measurements ", header_nms)
        mcorr_ = Obs(
            T=header_T, tmax=header_T - 1, nms=header_nms, is_resampled=resampled
        )
        # loop over file: read and store
        for indx, lndex in enumerate(file):
            # Read and store
            t = int(lndex.split(" ")[0])
            n = int(indx / header_T)
            # print(l.rstrip(), "     ", t, n)
            mcorr_.sample[n, t] = float(lndex.split(" ")[1])
        par.time_extent = header_T
        par.num_samples = header_nms
        par.assign_values()
        mcorr_.tmax = par.tmax
    #   Returns np array of correlators
    return mcorr_


def init_precision(digits_):
    mp.dps = digits_
    print(LogMessage(), "Setting precision ::::", "Binary precision in bit: ", mp.prec)
    print(
        LogMessage(),
        "Setting precision ::::",
        "Approximate decimal precision: ",
        mp.dps,
    )


class Inputs:
    def __init__(self):
        self.time_extent = -1
        self.tmax = 0
        self.datapath = "None"
        self.outdir = "None"
        self.logpath = "None"
        self.plotpath = "None"
        self.num_boot = 1
        self.num_samples = 1
        self.sigma = 0
        self.emax = -1
        self.Ne = 1
        self.Na = 1
        self.emin = 0
        self.e0 = 0
        self.periodicity = "EXP"
        self.A0cut = 0
        # self.l = -1
        self.prec = 105
        self.mpsigma = mpf("0")
        self.mpemax = mpf("0")
        self.mpemin = mpf("0")
        self.mpe0 = mpf("0")
        self.mplambda = mpf("0")
        self.directoryName = "."
        self.kerneltype = "FULLNORMGAUSS"
        self.loglevel = "WARNING"

    def _init(self, create_output_directories):
        init_precision(self.prec)
        if self.loglevel == "INFO":
            logger.setLevel(logging.INFO)
        elif self.loglevel == "DEBUG":
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.WARNING)
        if create_output_directories:
            self.plotpath, self.logpath = create_out_paths(self)

    def assign_values(self, create_output_directories=True):
        """
        Assigns tmax based on time_extent and periodicity if tmax was not specified
        Creates mpf(var) from float type var
        """
        if self.tmax == 0:
            if self.periodicity == "EXP":
                self.tmax = self.time_extent - 1  # Can't use c[0]
            elif self.periodicity == "COSH":
                self.tmax = int(
                    self.time_extent / 2
                )  # Can't use C[0] but can use c[T/2]
        if self.periodicity == "EXP":
            assert (self.tmax) < self.time_extent
        if self.periodicity == "COSH":
            assert (self.tmax) < self.time_extent / 2 + 1
        self.mpsigma = mpf(str(self.sigma))
        self.mpemax = mpf(str(self.emax))
        self.mpemin = mpf(str(self.emin))
        self.mpe0 = mpf(str(self.e0))
        self.directoryName = (
            "tmax"
            + str(self.tmax)
            + "sigma"
            + str(self.sigma)
            + "Ne"
            + str(self.Ne)
            + "nboot"
            + str(self.num_boot)
            + "prec"
            + str(self.prec)
            + "Na"
            + str(self.Na)
            + "KerType"
            + str(self.kerneltype)
        )
        self._init(create_output_directories)

    def report(self):
        print(LogMessage(), "Init ::: ", "Reading file:", self.datapath)
        print(LogMessage(), "Init ::: ", "Output directory:", self.outdir)
        print(LogMessage(), "Init ::: ", "Log directory:", self.logpath)
        print(LogMessage(), "Init ::: ", "Plot directory:", self.plotpath)
        print(LogMessage(), "Init ::: ", "Periodicity:", self.periodicity)
        print(LogMessage(), "Init ::: ", "Time extent:", self.time_extent)
        print(LogMessage(), "Init ::: ", "Smearing Kernel", self.kerneltype)
        print(LogMessage(), "Init ::: ", "tmax:", self.tmax)
        print(
            LogMessage(), "Init ::: ", "sigma (mp):", self.sigma, "(", self.mpsigma, ")"
        )
        print(LogMessage(), "Init ::: ", "Samples :", self.num_samples)
        print(LogMessage(), "Init ::: ", "Bootstrap samples :", self.num_boot)
        print(LogMessage(), "Init ::: ", "Number of energies :", self.Ne)
        print(
            LogMessage(),
            "Init ::: ",
            "Emax (mp)",
            self.emax,
            self.mpemax,
        )
        print(
            LogMessage(),
            "Init ::: ",
            "Emin (mp)",
            self.emin,
            "(",
            self.mpemin,
            ")",
        )
        print(LogMessage(), "Init ::: ", "Number of alphas", self.Na)
        print(LogMessage(), "Init ::: ", "Minimum value of A/A0 accepted ", self.A0cut)
        print(LogMessage(), "Init :::", "A integral from E0 = ", float(self.mpe0))


class MatrixBundle:
    """
    The B functional will be Bmatrix (the covariance, the identity, ...) normalised by bnorm (a number)
    """

    def __init__(self, Bmatrix: mp.matrix, bnorm=mpf(1)):
        self.B = Bmatrix
        self.bnorm = bnorm


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


CB_color_cycle = [
    "#377eb8",
    "#ff7f00",
    "#4daf4a",
    "#f781bf",
    "#a65628",
    "#984ea3",
    "#999999",
    "#e41a1c",
    "#dede00",
]

CB_colors = [
    "#1f77b4",  # Dark Blue
    "#ff7f0e",  # Orange
    "#2ca02c",  # Light Green
    "#d62728",  # Reddish Purple
    "#9467bd",  # Light Blue
    "#8c564b",  # Dark Yellow
    "#e377c2",  # Cyan
    "#7f7f7f",  # Olive Green
]

plot_markers = ["o", "s", "D", "v", "^", "p", "*", "h"]



from ..utils.rhoUtils import Obs, LogMessage, read_datafile
from ..utils.rhoParallelUtils import ParallelBootstrapLoop
from .correlatorUtils import foldPeriodicCorrelator, InputsCorrelatorAnalysis
from ..utils.rhoParser import parseArgumentCorrelatorAnalysis


def main():
    print(LogMessage(), "Initialising")
    args = parseArgumentCorrelatorAnalysis()
    par = InputsCorrelatorAnalysis(
        datapath=args.datapath, outdir=args.outdir, num_boot=args.nboot
    )

    #   Reading datafile, storing correlator
    rawcorr, par.time_extent, par.num_samples = read_datafile(par.datapath)
    par.periodicity = args.periodicity
    if args.periodicity == "EXP":
        par.tmax = par.time_extent - 1
    elif args.periodicity == "COSH":
        par.tmax = int(par.time_extent / 2) + 1
    else:
        raise ValueError("Invalid type specified. Only COSH and EXP are allowed.")

    par.report()

    #   Here is the correlator
    rawcorr.evaluate()
    #    rawcorr.plot(label="raw data")

    #   Here is the folding
    foldedCorr = foldPeriodicCorrelator(corr=rawcorr, par=par, is_resampled=False)
    foldedCorr.evaluate()
    #    foldedCorr.plot(label="folded")

    #   Here is the resampling
    corr = Obs(int(par.time_extent / 2) + 1, par.num_boot, is_resampled=True)
    resample = ParallelBootstrapLoop(par, foldedCorr.sample, is_folded=True)
    corr.sample = resample.run()
    corr.evaluate()
    corr.plot(show=True, label="Correlator (bootstrap)")

    print(LogMessage(), "Evaluate covariance")
    corr.evaluate_covmatrix(plot=False)
    corr.corrmat_from_covmat(plot=False)

    from correlatorUtils import effective_mass

    effmass = effective_mass(corr, par, type="COSH")
    effmass.plot(logscale=False)
    print(effmass.central, "±", effmass.err)

    exit(1)


if __name__ == "__main__":
    main()



from mpmath import mp, mpf
from .utils.rhoMath import cauchy
from scipy.integrate import quad
import numpy as np

def hlt_matrix(tmax: int, alpha, e0=mpf(0), type="EXP", T=0):
    S_ = mp.matrix(tmax, tmax)
    for i in range(tmax):
        for j in range(tmax):
            entry = mp.fadd(mpf(i), mpf(j))
            arg = mp.fadd(entry, mpf(2))  # i+j+2
            entry = mp.fsub(arg, alpha)  # i+j+2-a
            arg = mp.fneg(arg)
            arg = mp.fmul(arg, e0)
            arg = mp.exp(arg)
            entry = mp.fdiv(arg, entry)
            S_[i, j] = entry
            if type == "COSH":
                assert T > 0
                entry2 = mp.fsub(mpf(i), mpf(j))
                entry3 = mp.fsub(mpf(j), mpf(i))
                entry4 = mp.fneg(mpf(i))
                entry4 = mp.fsub(entry4, mpf(j))
                arg2 = mp.fadd(entry2, mpf(T))  # T+i-j
                arg3 = mp.fadd(entry3, mpf(T))  # T+j-i
                arg4 = mp.fadd(entry4, mpf(2 * T))  # 2T-j-i
                arg4 = mp.fsub(arg4, mpf(2))  # 2T-j-i-2
                entry2 = mp.fsub(arg2, alpha)  # T+i-j-a
                entry3 = mp.fsub(arg3, alpha)  # T+j-i-a
                entry4 = mp.fsub(arg4, alpha)  # 2T-j-i-2-a
                arg2 = mp.fneg(arg2)
                arg3 = mp.fneg(arg3)
                arg4 = mp.fneg(arg4)
                arg2 = mp.fmul(arg2, e0)
                arg3 = mp.fmul(arg3, e0)
                arg4 = mp.fmul(arg4, e0)
                arg2 = mp.exp(arg2)
                arg3 = mp.exp(arg3)
                arg4 = mp.exp(arg4)
                entry2 = mp.fdiv(arg2, entry2)
                entry3 = mp.fdiv(arg3, entry3)
                entry4 = mp.fdiv(arg4, entry4)
                S_[i, j] += entry2 + entry3 + entry4
    return S_


def generalised_ft(t, alpha, sigma, e, e0):
    newt = mp.fsub(t, alpha)  #
    aux = mp.fmul(sigma, sigma)  #   s^2
    arg = mp.fmul(aux, newt)  #   s^2 (t-alpha)
    aux = mp.fmul(arg, newt)  #   s^2 (alpha-t)^2
    aux = mp.fmul(aux, mpf("0.5"))  #   s^2 (alpha-t)^2 /2
    res = mp.exp(aux)  #   exp{s^2 (alpha-t)^2 /2}
    aux = mp.fneg(newt)  #   alpha-t
    aux = mp.fmul(e, aux)  #   e(alpha-t)
    aux = mp.exp(aux)
    res = mp.fmul(res, aux)  #   exp{s^2 (alpha-t)^2 /2} exp{estar (alpha-t) }
    #   The previous expression will multiply Erfc
    arg = mp.fadd(arg, e0)
    arg = mp.fsub(arg, e)  # s^2 (t-alpha) + E0 - omega
    arg = mp.fdiv(arg, sigma)
    aux = mp.sqrt(2)
    arg = mp.fdiv(arg, aux)  # [s^2 (t-alpha) + E0 - omega] / sqrt{2} s
    arg = mp.erfc(arg)  #   this is the COMPLEMENTARY erf
    res = mp.fmul(res, arg)  # Erfc() * exp()
    return res


def generalised_ft_halfnorm(t, alpha, sigma, e, e0):
    newt = mp.fsub(t, alpha)  #
    aux = mp.fmul(sigma, sigma)  # s^2
    arg = mp.fmul(aux, newt)  # s^2 (t-alpha)
    aux = mp.fmul(arg, newt)  # s^2 (alpha-t)^2
    aux = mp.fmul(aux, mpf("0.5"))  # s^2 (alpha-t)^2 /2
    res = mp.exp(aux)  # exp{s^2 (alpha-t)^2 /2}
    aux = mp.fneg(newt)  # alpha-t
    aux = mp.fmul(e, aux)  # e(alpha-t)
    aux = mp.exp(aux)
    res = mp.fmul(res, aux)  # exp{s^2 (alpha-t)^2 /2} exp{estar (alpha-t) }
    arg = mp.fadd(arg, e0)
    arg = mp.fsub(arg, e)
    arg = mp.fdiv(arg, sigma)
    aux = mp.sqrt(2)
    arg = mp.fdiv(arg, aux)
    arg = mp.erfc(arg)  # this is the COMPLEMENTARY erf
    res = mp.fmul(res, arg)
    aux = mp.fdiv(e, aux)
    aux = mp.fdiv(aux, sigma)
    aux = mp.erf(aux)
    aux = mp.fadd(mpf(1), aux)
    res = mp.fdiv(res, aux)
    return res


def ft_mp(e, t, sigma_, alpha, e0=mpf("0"), type="EXP", T=0, ker_type="FULLNORMGAUSS"):
    if ker_type == "FULLNORMGAUSS":
        res = generalised_ft(t, alpha, sigma_, e, e0)
        if type == "COSH":
            assert T > 0
            pterm = generalised_ft(T - t, alpha, sigma_, e, e0)
            res = mp.fadd(res, pterm)
        res *= 0.5

    elif ker_type == "HALFNORMGAUSS":
        res = generalised_ft_halfnorm(t, alpha, sigma_, e, e0)
        if type == "COSH":
            assert T > 0
            pterm = generalised_ft_halfnorm(T - t, alpha, sigma_, e, e0)
            res = mp.fadd(res, pterm)
    elif ker_type == "CAUCHY":
        # Define the function to be integrated
        def integrand(k):
            aux = mp.exp(alpha * k)
            aux2 = -t * k
            aux2 = mp.exp(aux2)
            if type == "COSH":
                assert T > 0
                aux3 = -(T - t) * k
                aux3 = mp.exp(aux3)
                aux2 = aux2 + aux3
            aux = aux * aux2 * cauchy(k, sigma_, e)
            return aux


        try:
            result, err = quad(integrand, float(e0), np.inf,
                               limit=200, epsabs=1e-10, epsrel=1e-10)
            #print('Error in integral: ', err)
            if not np.isfinite(result) or result < 0 or err > 1e-6:
                raise ValueError("Suspicious result")
            res = result
        except Exception as e:
            print(f"[SciPy fallback] Reason: {e}")
            res = mp.quad(integrand, [e0, mp.inf], method="gauss-legendre")


    else:
        raise ValueError("Invalid smearing kernel (par.ker_type)")
    return res


def a0_scalar(e, sigma, alpha, e0=mpf(0), ker_type="FULLNORMGAUSS"):
    """
    Functional A[g=0], for a given energy.
    """
    if ker_type == "FULLNORMGAUSS":
        aux = mp.fmul(sigma, sigma)  # start by the argument of erf
        aux = mp.fdiv(aux, mpf(2))  # s^2 / 2
        aux = mp.fmul(aux, alpha)  # a s^2 / 2
        aux = mp.fadd(e, aux)  # omega + a s^2 / 2
        aux = mp.fsub(aux, e0)  # omega - E0 + a s^2 / 2
        res = mp.fdiv(aux, sigma)  # omega - E0 + a s^2 / 2sigma
        res = mp.erf(res)  #   Erf
        res = mp.fadd(1, res)  # 1+erf, the numerator
        aux_ = mp.sqrt(mp.pi)
        res = mp.fdiv(res, aux_)  # 1+erf / sqrt{pi}
        res = mp.fdiv(res, sigma)  # 1+erf / (sqrt{pi} s)
        res = mp.fdiv(res, 4)  # 1+erf / (4 sqrt{pi} s)
        # exp term due to alpha
        if alpha != 0:
            aux = mp.fmul(alpha, e)  # alpha*e
            aux2 = mp.fmul(alpha, sigma)  # alpha*sigma
            aux2 = mp.fmul(aux2, aux2)  # (alpha*sigma)^2
            aux2 = mp.fdiv(aux2, mpf(4))  # (alpha*sigma)^2 / 4
            aux = mp.fadd(aux, aux2)  # (alpha*sigma)^2 / 4 + alphaomega
            aux = mp.exp(aux)
            res = mp.fmul(res, aux)
    elif ker_type == "HALFNORMGAUSS":
        aux = mp.fmul(sigma, sigma)
        aux = mp.fdiv(aux, mpf(2))
        aux = mp.fmul(aux, alpha)
        aux = mp.fadd(e, aux)
        aux = mp.fsub(aux, e0)
        res = mp.fdiv(aux, sigma)
        res = mp.erf(res)  #   Erf
        res = mp.fadd(1, res)  # 1+erf, the numerator
        aux_ = mp.sqrt(mp.pi)
        res = mp.fdiv(res, aux_)  # 1+erf /pi
        res = mp.fdiv(res, sigma)  # 1+erf / (sqrt{pi} s)
        aux_ = mp.sqrt(2)
        aux_ = mp.fdiv(e, aux_)
        aux_ = mp.fdiv(aux_, sigma)
        aux_ = mp.erf(aux_)
        aux_ = mp.fadd(aux_, 1)
        aux_ = mp.fmul(aux_, aux_)
        res = mp.fdiv(res, aux_)
        # alpha implementation
        aux = mp.fmul(alpha, e)  # alpha*e
        aux2 = mp.fmul(alpha, sigma)  # alpha*sigma
        aux2 = mp.fmul(aux2, aux2)  # (alpha*sigma)^2
        aux2 = mp.fdiv(aux2, mpf(4))  # (alpha*sigma)^2 / 4
        aux = mp.fadd(aux, aux2)  # (alpha*sigma)^2 / 4 + alpha*e
        aux = mp.exp(aux)
        res = mp.fmul(res, aux)
    elif ker_type == "CAUCHY":
        # Define the function to be integrated
        def integrand2(k):
            aux = alpha * k
            aux = mp.exp(aux)
            aux2 = cauchy(k, sigma, e) ** 2
            aux = aux * aux2
            return aux

        res = mp.quad(integrand2, [e0, mp.inf], method="gauss-legendre")
    else:
        raise ValueError("Invalid smearing kernel (par.ker_type)")

    return res


def a0_array(espace_mp, par, alpha):
    """
    Functional A[g=0], as an array of dimension len(espace_mp).
    """
    e0_ = par.e0
    a0_e = mp.matrix(par.Ne, 1)
    for ei in range(par.Ne):
        a0_e[ei] = a0_scalar(
            e=espace_mp[ei],
            sigma=par.mpsigma,
            alpha=mpf(alpha),
            e0=e0_,
            ker_type=par.kerneltype,
        )
    return a0_e


def integrandSigmaMat(e1, alpha, s, t1, t2, E0, par):
    _res = ft_mp(
        e=e1,
        t=t2,
        sigma_=s,
        alpha=alpha,
        e0=E0,
        type=par.periodicity,
        T=par.time_extent,
        ker_type=par.kerneltype,
    )
    if par.periodicity == "EXP":
        _res = _res * mp.exp(-t1 * e1)
    if par.periodicity == "COSH":
        _res = _res * (mp.exp(-t1 * e1) + mp.exp((-par.time_extent + t1) * e1))
    return _res


def SigmaMat(alpha, s, e0, par):
    SigmaMat_ = mp.matrix(par.tmax, par.tmax)

    for i in range(par.tmax):
        for j in range(par.tmax):
            entry = mp.quad(
                lambda x: integrandSigmaMat(x, alpha, s, i, j, e0, par),
                [e0, mp.inf],
                error=True,
            )
            SigmaMat_[i, j] = entry[0]
    return SigmaMat_


def gte(T, t, e, periodicity):
    """
    Laplace kernel exp(-t E) and its periodic generalisation.
    """
    if periodicity == "COSH":
        return mp.fadd(mp.exp((-T + t) * e), mp.exp(-t * e))
    if periodicity == "EXP":
        return mp.exp(-t * e)
